# -*- coding: utf-8 -*-
"""
Provide rules and states for a lattice-grain CTS model.

Created on Tue Feb  2 07:37:59 2016

@author: gtucker
"""

from landlab.ca.celllab_cts import Transition


def lattice_grain_node_states():
    """
    Create and return dict of states for lattice-grain model.
    """
    ns_dict = { 0 : 'empty', 
                1 : 'moving up',
                2 : 'moving right and up',
                3 : 'moving right and down',
                4 : 'moving down',
                5 : 'moving left and down',
                6 : 'moving left and up',
                7 : 'rest',
                8 : 'wall'}
    return ns_dict


def lattice_grain_transition_list(g=0.0, f=0.0, motion=1.0):
    """
    Creates and returns a list of Transition() objects to represent state
    transitions for simple granular mechanics model.
    
    Parameters
    ----------
    g : float (optional)
        Gravitational parameter. 1/g is the time taken for an initially
        stationary particle to fall one cell width. (Dimensions of 1/time)
    f : float (optional)
        Frictional parameter. Dimensions of 1/time. Probability per unit time
        of a frictional collision, in which one or both colliding particles
        come to a halt.
    motion : float (optional)
        Rate of motion (cells per time unit)

    Returns
    -------
    xn_list : list of Transition objects
        List of objects that encode information about the link-state transitions.

    Notes
    -----
    The transitions for this version of lattice gas have 11 pair-transition
    rules. The shorthand for the states is as follows:

        AR = air/empty
        IN = incoming particle (moving toward its neighbor)
        OU = outgoing particle (moving away from its neighbor)
        IO = incoming at an oblique angle
        OO = outgoing at an oblique angle
        RE = rest particle
        WA = wall particle
        op = oblique pair moving in opposite perpendicular direction
        sm = oblique pair moving in same perpendicular direction

    The 11 pairs with transitions are:

        1. AR-IN => IN-AR (move to empty particle)
        2. IN-IN => OO-OO-op (1/3 each dir), OU-OU (1/3) (head-on collision)
        3. IN-IO => OO-OU (oblique collision)
        4. IN-OO => IO-OU (oblique collision from behind)
        5. IN-OU => IO-OO (1/4 each of 2 directions) (collision from behind)
        6. IN-RE => RE-OU (1/3) RE-OO (1/3 each dir) (collision with rest)
        7. IN-WA => OU-WA (1/3) OO-WA (1/3 each dir) (wall collision)
        8. IO-IO-op => OO-OO-op (1/2 each dir) (glacing collision)
        9. IO-IO-sm => OO-OO-sm (30-degree collision)
        10. IO-RE => RE-OU (oblique collision with rest particle)
        11. IO-WA => OO-WA (oblique collision with wall)
    """
    xn_list = []

    # Set elastic-response rate, and scale frictional-response rate to motion
    # rate
    p_elast = motion * (1.0 - f)  # rate of elastic (non-dissipative) collision
    f *= motion

    # Rule 1: Transitions for particle movement into an empty cell
    xn_list.append( Transition((1,0,0), (0,1,0), motion, 'motion') )
    xn_list.append( Transition((2,0,1), (0,2,1), motion, 'motion') )
    xn_list.append( Transition((3,0,2), (0,3,2), motion, 'motion') )
    xn_list.append( Transition((0,4,0), (4,0,0), motion, 'motion') )
    xn_list.append( Transition((0,5,1), (5,0,1), motion, 'motion') )
    xn_list.append( Transition((0,6,2), (6,0,2), motion, 'motion') )

    # Rule 2: Transitions for head-on collision: elastic
    if p_elast > 0.0:
        xn_list.append( Transition((1,4,0), (4,1,0), p_elast/3, 'head-on collision') )
        xn_list.append( Transition((1,4,0), (3,6,0), p_elast/3, 'head-on collision') )
        xn_list.append( Transition((1,4,0), (5,2,0), p_elast/3, 'head-on collision') )
        xn_list.append( Transition((2,5,1), (5,2,1), p_elast/3, 'head-on collision') )
        xn_list.append( Transition((2,5,1), (4,1,1), p_elast/3, 'head-on collision') )
        xn_list.append( Transition((2,5,1), (6,3,1), p_elast/3, 'head-on collision') )
        xn_list.append( Transition((3,6,2), (6,3,2), p_elast/3, 'head-on collision') )
        xn_list.append( Transition((3,6,2), (1,4,2), p_elast/3, 'head-on collision') )
        xn_list.append( Transition((3,6,2), (5,2,2), p_elast/3, 'head-on collision') )

    # Rule 2: Transitions for head-on collision: frictional dissipation
    if f > 0.0:
        xn_list.append( Transition((1,4,0), (7,7,0), f, 'head-on collision') )
        xn_list.append( Transition((2,5,1), (7,7,1), f, 'head-on collision') )
        xn_list.append( Transition((3,6,2), (7,7,2), f, 'head-on collision') )

    # Rule 3: Transitions for oblique collision: elastic
    if p_elast > 0.0:
        xn_list.append( Transition((1,3,0), (3,1,0), p_elast, 'oblique collision') )
        xn_list.append( Transition((1,5,0), (5,1,0), p_elast, 'oblique collision') )
        xn_list.append( Transition((2,4,0), (4,2,0), p_elast, 'oblique collision') )
        xn_list.append( Transition((6,4,0), (4,6,0), p_elast, 'oblique collision') )
        xn_list.append( Transition((2,4,1), (4,2,1), p_elast, 'oblique collision') )
        xn_list.append( Transition((2,6,1), (6,2,1), p_elast, 'oblique collision') )
        xn_list.append( Transition((1,5,1), (5,1,1), p_elast, 'oblique collision') )
        xn_list.append( Transition((3,5,1), (5,3,1), p_elast, 'oblique collision') )
        xn_list.append( Transition((3,1,2), (1,3,2), p_elast, 'oblique collision') )
        xn_list.append( Transition((3,5,2), (5,3,2), p_elast, 'oblique collision') )
        xn_list.append( Transition((2,6,2), (6,2,2), p_elast, 'oblique collision') )
        xn_list.append( Transition((4,6,2), (6,4,2), p_elast, 'oblique collision') )

    # Rule 3 frictional
    if f > 0.0:
        xn_list.append( Transition((1,3,0), (7,7,0), f, 'oblique collision') )
        xn_list.append( Transition((1,5,0), (7,7,0), f, 'oblique collision') )
        xn_list.append( Transition((2,4,0), (7,7,0), f, 'oblique collision') )
        xn_list.append( Transition((6,4,0), (7,7,0), f, 'oblique collision') )
        xn_list.append( Transition((2,4,1), (7,7,1), f, 'oblique collision') )
        xn_list.append( Transition((2,6,1), (7,7,1), f, 'oblique collision') )
        xn_list.append( Transition((1,5,1), (7,7,1), f, 'oblique collision') )
        xn_list.append( Transition((3,5,1), (7,7,1), f, 'oblique collision') )
        xn_list.append( Transition((3,1,2), (7,7,2), f, 'oblique collision') )
        xn_list.append( Transition((3,5,2), (7,7,2), f, 'oblique collision') )
        xn_list.append( Transition((2,6,2), (7,7,2), f, 'oblique collision') )
        xn_list.append( Transition((4,6,2), (7,7,2), f, 'oblique collision') )

    # Rule 4: Transitions for oblique-from-behind collisions
    if p_elast > 0.0:
        xn_list.append( Transition((1,2,0), (2,1,0), p_elast, 'oblique') )
        xn_list.append( Transition((1,6,0), (6,1,0), p_elast, 'oblique') )
        xn_list.append( Transition((3,4,0), (4,3,0), p_elast, 'oblique') )
        xn_list.append( Transition((5,4,0), (4,5,0), p_elast, 'oblique') )
        xn_list.append( Transition((2,1,1), (1,2,1), p_elast, 'oblique') )
        xn_list.append( Transition((2,3,1), (3,2,1), p_elast, 'oblique') )
        xn_list.append( Transition((4,5,1), (5,4,1), p_elast, 'oblique') )
        xn_list.append( Transition((6,5,1), (5,6,1), p_elast, 'oblique') )
        xn_list.append( Transition((3,2,2), (2,3,2), p_elast, 'oblique') )
        xn_list.append( Transition((3,4,2), (4,3,2), p_elast, 'oblique') )
        xn_list.append( Transition((1,6,2), (6,1,2), p_elast, 'oblique') )
        xn_list.append( Transition((5,6,2), (6,5,2), p_elast, 'oblique') )

    # Rule 4 frictional
    if f > 0.0:
        xn_list.append( Transition((1,2,0), (7,1,0), f, 'oblique') )
        xn_list.append( Transition((1,6,0), (7,1,0), f, 'oblique') )
        xn_list.append( Transition((3,4,0), (4,7,0), f, 'oblique') )
        xn_list.append( Transition((5,4,0), (4,7,0), f, 'oblique') )
        xn_list.append( Transition((2,1,1), (7,2,1), f, 'oblique') )
        xn_list.append( Transition((2,3,1), (7,2,1), f, 'oblique') )
        xn_list.append( Transition((4,5,1), (5,7,1), f, 'oblique') )
        xn_list.append( Transition((6,5,1), (5,7,1), f, 'oblique') )
        xn_list.append( Transition((3,2,2), (7,3,2), f, 'oblique') )
        xn_list.append( Transition((3,4,2), (7,3,2), f, 'oblique') )
        xn_list.append( Transition((1,6,2), (6,7,2), f, 'oblique') )
        xn_list.append( Transition((5,6,2), (6,7,2), f, 'oblique') )
   
    # Rule 5: Transitions for direct-from-behind collisions
    if p_elast > 0.0:
        xn_list.append( Transition((1,1,0), (2,6,0), p_elast/4, 'behind') )
        xn_list.append( Transition((1,1,0), (6,2,0), p_elast/4, 'behind') )
        xn_list.append( Transition((4,4,0), (3,5,0), p_elast/4, 'behind') )
        xn_list.append( Transition((4,4,0), (5,3,0), p_elast/4, 'behind') )
        xn_list.append( Transition((2,2,1), (1,3,1), p_elast/4, 'behind') )
        xn_list.append( Transition((2,2,1), (3,1,1), p_elast/4, 'behind') )
        xn_list.append( Transition((5,5,1), (4,6,1), p_elast/4, 'behind') )
        xn_list.append( Transition((5,5,1), (6,4,1), p_elast/4, 'behind') )
        xn_list.append( Transition((3,3,2), (2,4,2), p_elast/4, 'behind') )
        xn_list.append( Transition((3,3,2), (4,2,2), p_elast/4, 'behind') )
        xn_list.append( Transition((6,6,2), (1,5,2), p_elast/4, 'behind') )
        xn_list.append( Transition((6,6,2), (5,1,2), p_elast/4, 'behind') )

    # Rule 5 frictional
    if f > 0.0:
        xn_list.append( Transition((1,1,0), (7,1,0), f/4, 'behind') )
        xn_list.append( Transition((4,4,0), (4,7,0), f/4, 'behind') )
        xn_list.append( Transition((2,2,1), (7,2,1), f/4, 'behind') )
        xn_list.append( Transition((5,5,1), (5,7,1), f/4, 'behind') )
        xn_list.append( Transition((3,3,2), (7,3,2), f/4, 'behind') )
        xn_list.append( Transition((6,6,2), (6,7,2), f/4, 'behind') )

    # Rule 6: Transitions for direct collision with stationary (resting) particle
    if p_elast > 0.0:
        xn_list.append( Transition((1,7,0), (7,1,0), p_elast/3., 'rest') )
        xn_list.append( Transition((1,7,0), (7,2,0), p_elast/3., 'rest') )
        xn_list.append( Transition((1,7,0), (7,6,0), p_elast/3., 'rest') )
        xn_list.append( Transition((7,4,0), (4,7,0), p_elast/3., 'rest') )
        xn_list.append( Transition((7,4,0), (3,7,0), p_elast/3., 'rest') )
        xn_list.append( Transition((7,4,0), (5,7,0), p_elast/3., 'rest') )
        xn_list.append( Transition((2,7,1), (7,2,1), p_elast/3., 'rest') )
        xn_list.append( Transition((2,7,1), (7,1,1), p_elast/3., 'rest') )
        xn_list.append( Transition((2,7,1), (7,3,1), p_elast/3., 'rest') )
        xn_list.append( Transition((7,5,1), (5,7,1), p_elast/3., 'rest') )
        xn_list.append( Transition((7,5,1), (4,7,1), p_elast/3., 'rest') )
        xn_list.append( Transition((7,5,1), (6,7,1), p_elast/3., 'rest') )
        xn_list.append( Transition((3,7,2), (7,3,2), p_elast/3., 'rest') )
        xn_list.append( Transition((3,7,2), (7,2,2), p_elast/3., 'rest') )
        xn_list.append( Transition((3,7,2), (7,4,2), p_elast/3., 'rest') )
        xn_list.append( Transition((7,6,2), (6,7,2), p_elast/3., 'rest') )
        xn_list.append( Transition((7,6,2), (1,7,2), p_elast/3., 'rest') )
        xn_list.append( Transition((7,6,2), (5,7,2), p_elast/3., 'rest') )

    # Rule 6 frictionl
    if f > 0.0:
        xn_list.append( Transition((1,7,0), (7,7,0), f, 'rest') )
        xn_list.append( Transition((7,4,0), (7,7,0), f, 'rest') )
        xn_list.append( Transition((2,7,1), (7,7,1), f, 'rest') )
        xn_list.append( Transition((7,5,1), (7,7,1), f, 'rest') )
        xn_list.append( Transition((3,7,2), (7,7,2), f, 'rest') )
        xn_list.append( Transition((7,6,2), (7,7,2), f, 'rest') )

    # Rule 7: Transitions for wall impact
    if p_elast > 0.0:
        xn_list.append( Transition((1,8,0), (4,8,0), p_elast/3, 'wall rebound') )
        xn_list.append( Transition((1,8,0), (3,8,0), p_elast/3, 'wall rebound') )
        xn_list.append( Transition((1,8,0), (5,8,0), p_elast/3, 'wall rebound') )
        xn_list.append( Transition((2,8,1), (5,8,1), p_elast/3, 'wall rebound') )
        xn_list.append( Transition((2,8,1), (4,8,1), p_elast/3, 'wall rebound') )
        xn_list.append( Transition((2,8,1), (6,8,1), p_elast/3, 'wall rebound') )
        xn_list.append( Transition((3,8,2), (6,8,2), p_elast/3, 'wall rebound') )
        xn_list.append( Transition((3,8,2), (5,8,2), p_elast/3, 'wall rebound') )
        xn_list.append( Transition((3,8,2), (1,8,2), p_elast/3, 'wall rebound') )
        xn_list.append( Transition((8,4,0), (8,1,0), p_elast/3, 'wall rebound') )
        xn_list.append( Transition((8,4,0), (8,6,0), p_elast/3, 'wall rebound') )
        xn_list.append( Transition((8,4,0), (8,2,0), p_elast/3, 'wall rebound') )
        xn_list.append( Transition((8,5,1), (8,1,1), p_elast/3, 'wall rebound') )
        xn_list.append( Transition((8,5,1), (8,2,1), p_elast/3, 'wall rebound') )
        xn_list.append( Transition((8,5,1), (8,3,1), p_elast/3, 'wall rebound') )
        xn_list.append( Transition((8,6,2), (8,2,2), p_elast/3, 'wall rebound') )
        xn_list.append( Transition((8,6,2), (8,3,2), p_elast/3, 'wall rebound') )
        xn_list.append( Transition((8,6,2), (8,4,2), p_elast/3, 'wall rebound') )

    # Rule 7 frictional
    if f > 0.0:
        xn_list.append( Transition((1,8,0), (7,8,0), f, 'wall rebound') )
        xn_list.append( Transition((2,8,1), (7,8,1), f, 'wall rebound') )
        xn_list.append( Transition((3,8,2), (7,8,2), f, 'wall rebound') )
        xn_list.append( Transition((8,4,0), (8,7,0), f, 'wall rebound') )
        xn_list.append( Transition((8,5,1), (8,7,1), f, 'wall rebound') )
        xn_list.append( Transition((8,6,2), (8,7,2), f, 'wall rebound') )

    # Rule 8: Transitions for glancing oblique collision
    if p_elast > 0.0:
        xn_list.append( Transition((2,5,0), (3,6,0), p_elast, 'glancing') )
        xn_list.append( Transition((6,3,0), (5,2,0), p_elast, 'glancing') )
        xn_list.append( Transition((3,6,1), (4,1,1), p_elast, 'glancing') )
        xn_list.append( Transition((1,4,1), (6,3,1), p_elast, 'glancing') )
        xn_list.append( Transition((4,1,2), (5,2,2), p_elast, 'glancing') )
        xn_list.append( Transition((2,5,2), (1,4,2), p_elast, 'glancing') )
    
    # Rule 8 frictional
    if f > 0.0:
        xn_list.append( Transition((2,5,0), (7,7,0), f, 'glancing') )
        xn_list.append( Transition((6,3,0), (7,7,0), f, 'glancing') )
        xn_list.append( Transition((3,6,1), (7,7,1), f, 'glancing') )
        xn_list.append( Transition((1,4,1), (7,7,1), f, 'glancing') )
        xn_list.append( Transition((4,1,2), (7,7,2), f, 'glancing') )
        xn_list.append( Transition((2,5,2), (7,7,2), f, 'glancing') )

    # Rule 9: Transitions for "near-on" collisions
    if p_elast > 0.0:
        xn_list.append( Transition((6,5,0), (5,6,0), p_elast, 'near-on') )
        xn_list.append( Transition((2,3,0), (3,2,0), p_elast, 'near-on') )
        xn_list.append( Transition((1,6,1), (6,1,1), p_elast, 'near-on') )
        xn_list.append( Transition((3,4,1), (4,3,1), p_elast, 'near-on') )
        xn_list.append( Transition((2,1,2), (1,2,2), p_elast, 'near-on') )
        xn_list.append( Transition((4,5,2), (5,4,2), p_elast, 'near-on') )
    
    # Rule 9 frictional
    if f > 0.0:
        xn_list.append( Transition((6,5,0), (7,6,0), f/2, 'near-on') )
        xn_list.append( Transition((6,5,0), (5,7,0), f/2, 'near-on') )
        xn_list.append( Transition((2,3,0), (7,2,0), f/2, 'near-on') )
        xn_list.append( Transition((2,3,0), (3,7,0), f/2, 'near-on') )
        xn_list.append( Transition((1,6,1), (7,1,1), f/2, 'near-on') )
        xn_list.append( Transition((1,6,1), (6,7,1), f/2, 'near-on') )
        xn_list.append( Transition((3,4,1), (7,3,1), f/2, 'near-on') )
        xn_list.append( Transition((3,4,1), (4,7,1), f/2, 'near-on') )
        xn_list.append( Transition((2,1,2), (7,2,2), f/2, 'near-on') )
        xn_list.append( Transition((2,1,2), (1,7,2), f/2, 'near-on') )
        xn_list.append( Transition((4,5,2), (7,4,2), f/2, 'near-on') )
        xn_list.append( Transition((4,5,2), (5,7,2), f/2, 'near-on') )
        
    # Rule 10: Transitions for oblique collision with rest particle
    if p_elast > 0.0:
        xn_list.append( Transition((2,7,0), (7,1,0), p_elast, 'oblique with rest') )
        xn_list.append( Transition((6,7,0), (7,1,0), p_elast, 'oblique with rest') )
        xn_list.append( Transition((7,3,0), (4,7,0), p_elast, 'oblique with rest') )
        xn_list.append( Transition((7,5,0), (4,7,0), p_elast, 'oblique with rest') )
        xn_list.append( Transition((3,7,1), (7,2,1), p_elast, 'oblique with rest') )
        xn_list.append( Transition((1,7,1), (7,2,1), p_elast, 'oblique with rest') )
        xn_list.append( Transition((7,6,1), (5,7,1), p_elast, 'oblique with rest') )
        xn_list.append( Transition((7,4,1), (5,7,1), p_elast, 'oblique with rest') )
        xn_list.append( Transition((4,7,2), (7,3,2), p_elast, 'oblique with rest') )
        xn_list.append( Transition((2,7,2), (7,3,2), p_elast, 'oblique with rest') )
        xn_list.append( Transition((7,5,2), (6,7,2), p_elast, 'oblique with rest') )
        xn_list.append( Transition((7,1,2), (6,7,2), p_elast, 'oblique with rest') )

    # Rule 10 frictional
    if f > 0.0:
        xn_list.append( Transition((2,7,0), (7,7,0), f, 'oblique with rest') )
        xn_list.append( Transition((6,7,0), (7,7,0), f, 'oblique with rest') )
        xn_list.append( Transition((7,3,0), (7,7,0), f, 'oblique with rest') )
        xn_list.append( Transition((7,5,0), (7,7,0), f, 'oblique with rest') )
        xn_list.append( Transition((3,7,1), (7,7,1), f, 'oblique with rest') )
        xn_list.append( Transition((1,7,1), (7,7,1), f, 'oblique with rest') )
        xn_list.append( Transition((7,6,1), (7,7,1), f, 'oblique with rest') )
        xn_list.append( Transition((7,4,1), (7,7,1), f, 'oblique with rest') )
        xn_list.append( Transition((4,7,2), (7,7,2), f, 'oblique with rest') )
        xn_list.append( Transition((2,7,2), (7,7,2), f, 'oblique with rest') )
        xn_list.append( Transition((7,5,2), (7,7,2), f, 'oblique with rest') )
        xn_list.append( Transition((7,1,2), (7,7,2), f, 'oblique with rest') )

    # Rule 11: Transitions for oblique collision with wall particle
    if p_elast > 0.0:
        xn_list.append( Transition((2,8,0), (3,8,0), p_elast, 'oblique with wall') )
        xn_list.append( Transition((6,8,0), (5,8,0), p_elast, 'oblique with wall') )
        xn_list.append( Transition((8,3,0), (8,2,0), p_elast, 'oblique with wall') )
        xn_list.append( Transition((8,5,0), (8,6,0), p_elast, 'oblique with wall') )
        xn_list.append( Transition((1,8,1), (6,8,1), p_elast, 'oblique with wall') )
        xn_list.append( Transition((3,8,1), (4,8,1), p_elast, 'oblique with wall') )
        xn_list.append( Transition((8,4,1), (8,3,1), p_elast, 'oblique with wall') )
        xn_list.append( Transition((8,6,1), (8,1,1), p_elast, 'oblique with wall') )
        xn_list.append( Transition((4,8,2), (5,8,2), p_elast, 'oblique with wall') )
        xn_list.append( Transition((2,8,2), (1,8,2), p_elast, 'oblique with wall') )
        xn_list.append( Transition((8,1,2), (8,2,2), p_elast, 'oblique with wall') )
        xn_list.append( Transition((8,5,2), (8,4,2), p_elast, 'oblique with wall') )

    # Rule 11 frictional
    if f > 0.0:
        xn_list.append( Transition((2,8,0), (7,8,0), f, 'oblique with wall') )
        xn_list.append( Transition((6,8,0), (7,8,0), f, 'oblique with wall') )
        xn_list.append( Transition((8,3,0), (8,7,0), f, 'oblique with wall') )
        xn_list.append( Transition((8,5,0), (8,7,0), f, 'oblique with wall') )
        xn_list.append( Transition((1,8,1), (7,8,1), f, 'oblique with wall') )
        xn_list.append( Transition((3,8,1), (7,8,1), f, 'oblique with wall') )
        xn_list.append( Transition((8,4,1), (8,7,1), f, 'oblique with wall') )
        xn_list.append( Transition((8,6,1), (8,7,1), f, 'oblique with wall') )
        xn_list.append( Transition((4,8,2), (7,8,2), f, 'oblique with wall') )
        xn_list.append( Transition((2,8,2), (7,8,2), f, 'oblique with wall') )
        xn_list.append( Transition((8,1,2), (8,7,2), f, 'oblique with wall') )
        xn_list.append( Transition((8,5,2), (8,7,2), f, 'oblique with wall') )

    # Gravity rule 1: rising particles become rest particles
    if g > 0.0:
        xn_list.append( Transition((0,1,0), (0,7,0), g, 'gravity 1') )
        xn_list.append( Transition((1,1,0), (1,7,0), g, 'gravity 1') )
        xn_list.append( Transition((2,1,0), (2,7,0), g, 'gravity 1') )
        xn_list.append( Transition((3,1,0), (3,7,0), g, 'gravity 1') )
        xn_list.append( Transition((4,1,0), (4,7,0), g, 'gravity 1') )
        xn_list.append( Transition((5,1,0), (5,7,0), g, 'gravity 1') )
        xn_list.append( Transition((6,1,0), (6,7,0), g, 'gravity 1') )
        xn_list.append( Transition((7,1,0), (7,7,0), g, 'gravity 1') )
        xn_list.append( Transition((8,1,0), (8,7,0), g, 'gravity 1') )

    # Gravity rule 2: resting particles become falling particles (if not above
    # rest or wall?)
    if g > 0.0:
        xn_list.append( Transition((0,7,0), (0,4,0), g, 'gravity 2') )
#        xn_list.append( Transition((1,7,0), (1,4,0), g, 'gravity 2') )
#        xn_list.append( Transition((2,7,0), (2,4,0), g, 'gravity 2') )
#        xn_list.append( Transition((3,7,0), (3,4,0), g, 'gravity 2') )
#        xn_list.append( Transition((4,7,0), (4,4,0), g, 'gravity 2') )
#        xn_list.append( Transition((5,7,0), (5,4,0), g, 'gravity 2') )
#        xn_list.append( Transition((6,7,0), (6,4,0), g, 'gravity 2') )

    # Gravity rule 3: up/sideways particles become down/sideways particles
    if g > 0.0:
        xn_list.append( Transition((0,2,0), (0,3,0), g, 'gravity 3') )
        xn_list.append( Transition((1,2,0), (1,3,0), g, 'gravity 3') )
        xn_list.append( Transition((2,2,0), (2,3,0), g, 'gravity 3') )
        xn_list.append( Transition((3,2,0), (3,3,0), g, 'gravity 3') )
        xn_list.append( Transition((4,2,0), (4,3,0), g, 'gravity 3') )
        xn_list.append( Transition((5,2,0), (5,3,0), g, 'gravity 3') )
        xn_list.append( Transition((6,2,0), (6,3,0), g, 'gravity 3') )
        xn_list.append( Transition((7,2,0), (7,3,0), g, 'gravity 3') )
        xn_list.append( Transition((8,2,0), (8,3,0), g, 'gravity 3') )
        xn_list.append( Transition((0,6,0), (0,5,0), g, 'gravity 3') )
        xn_list.append( Transition((1,6,0), (1,5,0), g, 'gravity 3') )
        xn_list.append( Transition((2,6,0), (2,5,0), g, 'gravity 3') )
        xn_list.append( Transition((3,6,0), (3,5,0), g, 'gravity 3') )
        xn_list.append( Transition((4,6,0), (4,5,0), g, 'gravity 3') )
        xn_list.append( Transition((5,6,0), (5,5,0), g, 'gravity 3') )
        xn_list.append( Transition((6,6,0), (6,5,0), g, 'gravity 3') )
        xn_list.append( Transition((7,6,0), (7,5,0), g, 'gravity 3') )
        xn_list.append( Transition((8,6,0), (8,5,0), g, 'gravity 3') )
    
    # Gravity rule 4: down/side to straight down
    if g > 0.0:
        xn_list.append( Transition((0,3,0), (0,4,0), g, 'gravity 4') )
        xn_list.append( Transition((1,3,0), (1,4,0), g, 'gravity 4') )
        xn_list.append( Transition((2,3,0), (2,4,0), g, 'gravity 4') )
        xn_list.append( Transition((3,3,0), (3,4,0), g, 'gravity 4') )
        xn_list.append( Transition((4,3,0), (4,4,0), g, 'gravity 4') )
        xn_list.append( Transition((5,3,0), (5,4,0), g, 'gravity 4') )
        xn_list.append( Transition((6,3,0), (6,4,0), g, 'gravity 4') )
        xn_list.append( Transition((7,3,0), (7,4,0), g, 'gravity 4') )
        xn_list.append( Transition((8,3,0), (8,4,0), g, 'gravity 4') )
        xn_list.append( Transition((0,5,0), (0,4,0), g, 'gravity 4') )
        xn_list.append( Transition((1,5,0), (1,4,0), g, 'gravity 4') )
        xn_list.append( Transition((2,5,0), (2,4,0), g, 'gravity 4') )
        xn_list.append( Transition((3,5,0), (3,4,0), g, 'gravity 4') )
        xn_list.append( Transition((4,5,0), (4,4,0), g, 'gravity 4') )
        xn_list.append( Transition((5,5,0), (5,4,0), g, 'gravity 4') )
        xn_list.append( Transition((6,5,0), (6,4,0), g, 'gravity 4') )
        xn_list.append( Transition((7,5,0), (7,4,0), g, 'gravity 4') )
        xn_list.append( Transition((8,5,0), (8,4,0), g, 'gravity 4') )

    # Gravity rule for lateral destabilization (represents grain
    # motion above angle of repose on sloping surface)
    if g > 0.0:
        xn_list.append( Transition((7,0,2), (3,0,2), g/2.0, 'gravity') )
        xn_list.append( Transition((0,7,1), (0,5,1), g/2.0, 'gravity') )
 
    return xn_list   



name = 'landlab'
cite_as = [
    """@article{hobley2017creative,
    title={Creative computing with Landlab: an open-source toolkit
        for building, coupling, and exploring two-dimensional
        numerical models of Earth-surface dynamics},
    author={Hobley, Daniel EJ and Adams, Jordan M and Nudurupati,
        Sai Siddhartha and Hutton, Eric WH and Gasparini, Nicole M and
        Istanbulluoglu, Erkan and Tucker, Gregory E},
    journal={Earth Surface Dynamics},
    volume={5},
    number={1},
    pages={21},
    year={2017},
    publisher={Copernicus GmbH}
    }""",
]



#! /usr/bin/env python
"""
Examples
--------
>>> import numpy as np
>>> from landlab.grid.structured_quad.structured import StructuredQuadGrid
>>> (y, x) = np.meshgrid(np.arange(4.), np.arange(5.), indexing='ij')
>>> grid = StructuredQuadGrid((y, x))
>>> grid.number_of_nodes
20

>>> grid.number_of_node_rows == 4
True
>>> grid.number_of_node_columns == 5
True
>>> grid.nodes_at_corners_of_grid
array([ 0,  4, 15, 19])
>>> grid.number_of_cells
6
"""
import numpy as np

from ..base import FIXED_VALUE_BOUNDARY
from ..unstructured.status import StatusGrid
from ..unstructured.base import BaseGrid
from .links import (StructuredQuadLinkGrid, node_id_at_link_start,
                    node_id_at_link_end)
from .cells import StructuredQuadCellGrid
from . import cells as quad_cells
from . import links as quad_links
from . import faces as quad_faces
from . import nodes
from landlab.utils.decorators import deprecated


class StructuredQuadGrid(BaseGrid):
    """
    Parameters
    ----------
    node_coord : tuple
        Coordinates of all grid nodes.
    shape : tuple, optional
        Shape of the grid of nodes.
    """

    def __init__(self, node_coord, shape=None, axis_name=None, axis_units=None,
                 links=True, cells=True, node_status=None):
        """
        Parameters
        ----------
        node_coord : tuple
            Coordinates of all grid nodes.
        shape : tuple, optional
            Shape of the grid of nodes.
        """
        if len(node_coord) != 2:
            raise ValueError('only 2d grids are supported')

        self._shape = shape or node_coord[0].shape

        if node_status is not None:
            if node_status.size != nodes.number_of_nodes(self.shape):
                raise ValueError('incorrect size for node_status array')

        if node_status is None:
            self._status = nodes.status_with_perimeter_as_boundary(
                self.shape, node_status=FIXED_VALUE_BOUNDARY)
        else:
            self._status = node_status

        if links:
            # links = (node_id_at_link_start(self.shape),
            #         node_id_at_link_end(self.shape))
            link_grid = StructuredQuadLinkGrid(self.shape)
        if cells:
            cell_grid = StructuredQuadCellGrid(self.shape)

        #super(StructuredQuadGrid, self).__init__(node_status=node_status)
        BaseGrid.__init__(self, (node_coord[0].flatten(),
                                 node_coord[1].flatten()),
                          links=link_grid,
                          cells=cell_grid)

        self._num_nodes = nodes.number_of_nodes(self.shape)
        self._num_cells = quad_cells.number_of_cells(self.shape)
        self._num_links = quad_links.number_of_links(self.shape)
        self._num_faces = quad_faces.number_of_faces(self.shape)

        self._num_core_nodes = nodes.number_of_core_nodes(self.shape)
        self._num_core_cells = self._num_cells

        self._node_x, self._node_y = (
            np.ravel(node_coord[0]),
            np.ravel(node_coord[1]),
        )

        self._node_id_at_cells = quad_cells.node_id_at_cells(self.shape)
        self._cell_id_at_nodes = quad_cells.cell_ids(self.shape)

        self._cell_node = quad_cells.node_id_at_cells(self.shape)

        self._in_link_id_at_nodes = quad_links.node_in_link_ids(self.shape)
        self._out_link_id_at_nodes = quad_links.node_out_link_ids(self.shape)

        self._node_id_at_link_start = quad_links.node_id_at_link_start(
            self.shape)
        self._node_id_at_link_end = quad_links.node_id_at_link_end(self.shape)

        self._active_link_ids = quad_links.active_link_ids(
            self.shape, self._status)

    @property
    def shape(self):
        """Shape of the grid as rows, columns.
        """
        return self._shape

    #@property
    # def number_of_core_nodes(self):
    #    """Number of core nodes.
    #    """
    #    return self._num_core_nodes

    @property
    def number_of_node_columns(self):
        """Number of node columns.

        Returns the number of columns, including boundaries.
        """
        return self.shape[1]

    @property
    def number_of_node_rows(self):
        """Number of node rows.

        Returns the number of rows, including boundaries.
        """
        return self.shape[0]

    @property
    @deprecated(use='nodes_at_corners_of_grid', version=1.0)
    def corner_nodes(self):
        return self.nodes_at_corners_of_grid

    @property
    def nodes_at_corners_of_grid(self):
        """Nodes in grid corners.

        Return the IDs to the corner nodes of the grid, sorted by ID.

        Returns
        -------
        (4, ) ndarray
            Array of corner node IDs.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab.grid.structured_quad.structured import StructuredQuadGrid
        >>> (x, y) = np.meshgrid(np.arange(4.), np.arange(5.), indexing='ij')
        >>> grid = StructuredQuadGrid((x, y))
        >>> grid.nodes_at_corners_of_grid
        array([ 0,  4, 15, 19])
        """
        return nodes.corners(self.shape)



import numpy as np
from numpy.testing import assert_array_equal

from landlab import FIXED_GRADIENT_BOUNDARY, CLOSED_BOUNDARY, INACTIVE_LINK, \
    FIXED_LINK
from landlab import RasterModelGrid


def test_issue_428_a():
    """Issue #428"""
    grid = RasterModelGrid((4, 4))
    grid.set_closed_boundaries_at_grid_edges(True, True, True, True)

    assert grid.status_at_node[1] == 4
    assert grid.status_at_link[4] == 4
    assert_array_equal(grid.active_link_dirs_at_node[1],[0, 0, 0, 0])

    grid.status_at_node[1] = 1
    assert grid.status_at_link[4] == 0
    assert_array_equal(grid.active_link_dirs_at_node[1], [0, -1, 0, 0])


def test_issue_428_b():
    """Issue #428"""
    grid = RasterModelGrid((4, 4))

    z = np.ones(grid.number_of_nodes)
    z[grid.nodes_at_bottom_edge] = -9999.
    z[grid.nodes_at_left_edge] = -9999.
    z[grid.nodes_at_top_edge] = -9999.
    z[grid.nodes_at_right_edge] = -9999.
    z[1] = .5

    assert_array_equal(grid.active_link_dirs_at_node[1], [0, -1, 0, 0])

    grid.set_watershed_boundary_condition(z)
    assert_array_equal(grid.active_link_dirs_at_node[1], [0, -1, 0, 0])


def test_link_update_with_nodes_closed():
    rmg = RasterModelGrid((4, 5))
    rmg.status_at_node[rmg.nodes_at_bottom_edge] = CLOSED_BOUNDARY
    inactive_array = np.array([INACTIVE_LINK, ] * 5)
    assert_array_equal(rmg.status_at_link[4:9], inactive_array)


def test_link_update_with_nodes_fixed_grad():
    rmg = RasterModelGrid((4, 5))
    rmg.status_at_node[rmg.nodes_at_bottom_edge] = FIXED_GRADIENT_BOUNDARY
    fixed_array = np.array([FIXED_LINK, ] * 3)
    assert_array_equal(rmg.status_at_link[5:8], fixed_array)


def test_bc_set_code_init():
    rmg = RasterModelGrid((4, 5))
    assert rmg.bc_set_code == 0


def test_bc_set_code_change():
    rmg = RasterModelGrid((4, 5))
    rmg.status_at_node[rmg.nodes_at_bottom_edge] = CLOSED_BOUNDARY
    assert rmg.bc_set_code != 0



import numpy as np
import xarray as xr

import six

from .scalar_data_fields import FieldError
from .grouped import GroupError


FIELD_LOCATIONS = ('node', 'link', 'patch', 'corner', 'face', 'cell', )


def reshape_for_storage(array, field_size=None):
    """Reshape an array to be stored as a field.

    For reshaping rules, see ``shape_for_storage``.

    Parameters
    ----------
    array : ndarray
        The array to be stored.
    field_size : int, optional
        The size of the field.

    Returns
    -------
    ndarray
        The (possibly) reshaped array.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.field.graph_field import reshape_for_storage

    The shape will be such that the first dimension in the field size.

    >>> data = np.arange(6)
    >>> reshape_for_storage(data, 3)
    array([[0, 1],
           [2, 3],
           [4, 5]])
    >>> reshape_for_storage(data, 2)
    array([[0, 1, 2],
           [3, 4, 5]])
    >>> reshape_for_storage(data, 6)
    array([0, 1, 2, 3, 4, 5])

    If the array is already the correct shape, just return that array.

    >>> data = np.arange(6).reshape((2, 3))
    >>> reshape_for_storage(data, 2) is data
    True
    """
    shape = shape_for_storage(array, field_size)
    if shape == array.shape or array.ndim == 0:
        return array
    else:
        return array.reshape(shape)


def shape_for_storage(array, field_size=None):
    """The shape an array will be stored as.

    Parameters
    ----------
    array : ndarray
        The array to be stored.
    field_size : int, optional
        The size of the field.

    Returns
    -------
    tuple of int
        The shape the array will be stored as.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.field.graph_field import shape_for_storage

    The shape will be such that the first dimension in the field size.

    >>> data = np.arange(6)
    >>> shape_for_storage(data, 3) == (3, 2)
    True
    >>> shape_for_storage(data, 2) == (2, 3)
    True
    >>> shape_for_storage(data, 6) == (6, )
    True

    If a field size is not given, the array will be stored as a
    flattened array.

    >>> shape_for_storage(data) == (6, )
    True
    >>> data = np.arange(6).reshape((3, 2))
    >>> shape_for_storage(data) == (6, )
    True

    For field sizes of 1, the array is always flattened.

    >>> shape_for_storage(data, 1) == (6, )
    True

    For scalar arrays, the field size must be 1.

    >>> data = np.array(1.)
    >>> shape_for_storage(data) == (1, )
    True
    >>> shape_for_storage(data, field_size=1) == (1, )
    True

    If the array cannot be shaped into a storage shape, a `ValueError`
    is raised.

    >>> data = np.array(1.)
    >>> shape_for_storage(data, field_size=4) # DOCTEST: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    ValueError: unable to reshape array to field size
    >>> data = np.arange(6.)
    >>> shape_for_storage(data, field_size=4) # DOCTEST: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    ValueError: unable to reshape array to field size
    """
    if field_size is None:
        field_size = array.size

    if array.size % field_size != 0:
        raise ValueError('unable to reshape array to field size')

    if field_size in (1, array.size):
        shape = (array.size, )
    else:
        shape = (field_size, array.size // field_size)

    return shape


class FieldDataset(dict):

    """Wrap an xarray.Dataset as a landlab field.

    This is a light wrapping of xarray.Dataset. The main differences
    are that a `FieldDataset` can be created with a size but not
    allocate any memory for data arrays until an array is actually
    needed. The setitem method is also overriden so that when arrays
    are added they are stored reshaped in the landlab style. That
    is shaped as `(n_elements, values_per_element)`.
    """

    def __init__(self, *args, **kwds):
        self._name, self._size = args[0], args[1]
        self._ds = xr.Dataset()
        self._units = {}

    @property
    def size(self):
        return self._size

    @property
    def units(self):
        return self._units

    @property
    def dataset(self):
        return self._ds

    def keys(self):
        return self._ds.variables

    def set_value(self, name, value_array, attrs=None):
        attrs = attrs or {}
        attrs.setdefault('units', '?')

        value_array = np.asarray(value_array)

        if not self._size:
            self._size = value_array.size

        if name in self._ds and self._ds[name].values is value_array:
            self._ds[name].values.shape = shape_for_storage(value_array,
                                                            self.size)
            return

        value_array = reshape_for_storage(value_array, self._size)

        if self._size == 1:
            if value_array.ndim > 0:
                dims = (name + '_per_' + self._name, )
            else:
                dims = ()
        else:
            dims = (self._name, )
            if value_array.ndim > 1:
                dims += (name + '_per_' + self._name, )

        if name in self._ds:
            self._ds = self._ds.drop(name)

        self._ds.update({name: xr.DataArray(value_array, dims=dims,
                                            attrs=attrs)})
        self._units[name] = attrs['units']

    def __getitem__(self, name):
        if isinstance(name, six.string_types):
            try:
                return self._ds[name].values
            except KeyError:
                raise FieldError(name)
        else:
            raise TypeError('field name not a string')

    def __setitem__(self, name, value_array):
        self.set_value(name, value_array)

    def __contains__(self, name):
        return name in self._ds

    def __str__(self):
        return str(self._ds)

    # def __repr__(self):
    #     return repr(self._ds)

    def __len__(self):
        return self._size


class GraphFields(object):

    """Collection of grouped data-fields.

    The `GraphFields` class holds a set of data fields that are
    separated into *groups*. A typical use for this class would be to define
    the groups as being locations on a grid where the values are defined.
    For instance, the groups could be *node*, *cell*, *link*, and *face*.

    Attributes
    ----------
    groups

    See Also
    --------
    landlab.field.ScalarDataFields : Data fields within a *group* are
        stored as :class:`landlab.field.ScalarDataFields`.
    landlab.field.ModelDataFields : Equivalent data structure for
        old-style fields.

    Examples
    --------

    Create two groups of data fields defined at *node* and *cell*. Each set can
    have a differenct number of values.

    >>> from landlab.field import GraphFields
    >>> fields = GraphFields()
    >>> fields.new_field_location('node', 12)
    >>> fields.new_field_location('cell', 2)

    Create some new value arrays for each of the data fields.

    >>> fields.ones('node')
    array([ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.])
    >>> fields.zeros('cell')
    array([ 0.,  0.])

    Create new value arrays and add them to the data fields. Because the data
    fields are in different groups (node and cell), they can have the same
    name.

    >>> fields.add_ones('node', 'topographic__elevation')
    array([ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.])
    >>> fields.at_node['topographic__elevation']
    array([ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.])

    >>> fields.add_ones('cell', 'topographic__elevation')
    array([ 1.,  1.])
    >>> fields.at_cell['topographic__elevation']
    array([ 1.,  1.])

    Each group acts as a `dict` so, for instance, to get the variables names
    in a group use the `keys` method,

    >>> list(fields.at_cell.keys())
    ['topographic__elevation']
    """

    def __init__(self, *args, **kwds):
        try:
            dims = args[0]
        except IndexError:
            dims = {}

        self._groups = set()
        for loc in FIELD_LOCATIONS:
            if loc in dims:
                self.new_field_location(loc, dims[loc])
            else:
                setattr(self, '_at_' + loc, None)

        self._default_group = 'node'

    def __getitem__(self, name):
        try:
            return getattr(self, 'at_' + name)
        except AttributeError:
            raise GroupError(name)

    @property
    def default_group(self):
        return self._default_group

    @default_group.setter
    def default_group(self, loc):
        if self.has_group(loc):
            self._defualt_group = loc
        else:
            raise ValueError('{loc} is not a valid group name'.format(loc=loc))

    def new_field_location(self, loc, size=None):
        """Add a new quantity to a field.

        Create an empty group into which new fields can be added. The new group
        is created but no memory allocated yet. The dictionary of the new group
        can be through a new *at_* attribute of the class instance.

        Parameters
        ----------
        group: str
            Name of the new group to add to the field.
        size: int, optional
            Number of elements in the new quantity. If not provided, the
            size is set to be the size of the first field added to the goup.

        Raises
        ------
        ValueError
            If the field already contains the group.

        Examples
        --------
        Create a collection of fields and add two groups, *node* and *cell*,
        to it.

        >>> from landlab.field import GraphFields
        >>> fields = GraphFields()
        >>> fields.new_field_location('node', 12)
        >>> fields.new_field_location('cell', 2)

        The group names in the collection are retrieved with the *groups*
        attribute as a `set`.

        >>> names = list(fields.groups)
        >>> names.sort()
        >>> names
        ['cell', 'node']

        Access the new (empty) groups with the *at_* attributes.

        >>> fields.at_cell, fields.at_node
        ({}, {})

        >>> fields.new_field_location('core_node')
        >>> fields.at_core_node.size is None
        True
        >>> fields.at_core_node['air__temperature'] = [0, 1]
        >>> fields.at_core_node.size
        2

        LLCATS: FIELDCR
        """
        dataset_name = 'at_' + loc
        if loc not in self._groups:
            setattr(self, dataset_name, FieldDataset(loc, size))
            self._groups.add(loc)
        else:
            raise ValueError('{loc} location already exists'.format(loc=loc))

    @property
    def groups(self):
        """List of group names.

        Returns
        -------
        set
            Set of quantity names.
        """
        return self._groups

    def has_group(self, name):
        """Check if a group exists.

        Parameters
        ----------
        group: str
            Name of the group.

        Returns
        -------
        boolean
            True if the field contains *group*, otherwise False.

        Examples
        --------
        Check if the field has the groups named *node* or *cell*.

        >>> from landlab.field import GraphFields
        >>> fields = GraphFields()
        >>> fields.new_field_location('node', 12)
        >>> fields.has_group('node')
        True
        >>> fields.has_group('cell')
        False

        LLCATS: FIELDINF
        """
        return name in self._groups

    def has_field(self, group, field):
        """Check if a field is in a group.

        Parameters
        ----------
        group: str
            Name of the group.
        field: str
            Name of the field.

        Returns
        -------
        boolean
            ``True`` if the group contains the field, otherwise ``False``.

        Examples
        --------
        Check if the field named ``topographic__elevation`` is contained
        in a group.

        >>> from landlab.field import GraphFields
        >>> fields = GraphFields()
        >>> fields.new_field_location('node', 12)
        >>> _ = fields.add_ones('node', 'topographic__elevation')
        >>> fields.has_field('node', 'topographic__elevation')
        True
        >>> fields.has_field('cell', 'topographic__elevation')
        False

        LLCATS: FIELDINF
        """
        try:
            return field in self[group]
        except KeyError:
            return False

    def keys(self, group):
        """List of field names in a group.

        Returns a list of the field names as a list of strings.

        Parameters
        ----------
        group : str
            Group name.

        Returns
        -------
        list
            List of field names.

        Examples
        --------
        >>> from landlab.field import GraphFields
        >>> fields = GraphFields()
        >>> fields.new_field_location('node', 4)
        >>> list(fields.keys('node'))
        []
        >>> _ = fields.add_empty('node', 'topographic__elevation')
        >>> list(fields.keys('node'))
        ['topographic__elevation']

        LLCATS: FIELDINF
        """
        return self[group].keys()

    def size(self, group):
        """Size of the arrays stored in a group.

        Parameters
        ----------
        group : str
            Group name.

        Returns
        -------
        int
            Array size.

        Examples
        --------
        >>> from landlab.field import GraphFields
        >>> fields = GraphFields()
        >>> fields.new_field_location('node', 4)
        >>> fields.size('node')
        4

        LLCATS: GINF FIELDINF
        """
        return len(self[group])

    def field_values(self, group, field):
        """Get values of a field.

        Given a *group* and a *field*, return a reference to the associated
        data array.

        Parameters
        ----------
        group: str
            Name of the group.
        field: str
            Name of the field withing *group*.

        Returns
        -------
        array
            The values of the field.

        Raises
        ------
        GroupError
            If *group* does not exits
        FieldError
            If *field* does not exits

        Examples
        --------
        Create a group of fields called *node*.

        >>> from landlab.field import GraphFields
        >>> fields = GraphFields()
        >>> fields.new_field_location('node', 4)

        Add a field, initialized to ones, called *topographic__elevation*
        to the *node* group. The *field_values* method returns a reference
        to the field's data.

        >>> _ = fields.add_ones('node', 'topographic__elevation')
        >>> fields.field_values('node', 'topographic__elevation')
        array([ 1.,  1.,  1.,  1.])

        Raise FieldError if *field* does not exist in *group*.

        >>> fields.field_values('node', 'planet_surface__temperature')
        ...     # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        FieldError: planet_surface__temperature

        If *group* does not exists, Raise GroupError.

        >>> fields.field_values('cell', 'topographic__elevation')
        ...     # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        GroupError: cell

        LLCATS: FIELDIO
        """
        try:
            fields = self[group]
        except KeyError:
            raise GroupError(group)
        try:
            return fields[field]
        except KeyError:
            raise FieldError(group)

    def field_units(self, group, field):
        """Get units for a field.

        Returns the unit string associated with the data array in *group* and
        *field*.

        Parameters
        ----------
        group: str
            Name of the group.
        field: str
            Name of the field withing *group*.

        Returns
        -------
        str
            The units of the field.

        Raises
        ------
        KeyError
            If either *field* or *group* does not exist.

        LLCATS: FIELDINF
        """
        return self[group]._ds[field].attrs['units']

    def empty(self, *args, **kwds):
        """Uninitialized array whose size is that of the field.

        Return a new array of the data field size, without initializing
        entries. Keyword arguments are the same as that for the equivalent
        numpy function.

        Parameters
        ----------
        group : str
            Name of the group.

        See Also
        --------
        numpy.empty : See for a description of optional keywords.
        landlab.field.GraphFields.ones : Equivalent method that
            initializes the data to 1.
        landlab.field.GraphFields.zeros : Equivalent method that
            initializes the data to 0.

        Examples
        --------
        >>> from landlab.field import GraphFields
        >>> field = GraphFields()
        >>> field.new_field_location('node', 4)
        >>> field.empty('node') # doctest: +SKIP
        array([  2.31584178e+077,  -2.68156175e+154,   9.88131292e-324,
        ... 2.78134232e-309]) # Uninitialized memory

        Note that a new field is *not* added to the collection of fields.

        >>> list(field.keys('node'))
        []

        LLCATS: FIELDCR
        """
        if len(args) == 0:
            group = kwds.pop('at', kwds.pop('centering', 'node'))
        else:
            group = args[0]
        
        if group == 'grid':
            raise ValueError("ones is not supported for at='grid', if you "
                             "want to create a field at the grid, use\n"
                             "grid.at_grid['value_name']=value\n"
                             "instead.\nAlternatively, if you want ones"
                             "of the shape stored at_grid, use np.array(1).")
        try:
            size = len(getattr(self, 'at_' + group))
        except AttributeError:
            size = getattr(self, 'number_of_' + group)

        return np.empty(size, **kwds)

    def ones(self, *args, **kwds):
        """Array, initialized to 1, whose size is that of the field.

        Return a new array of the data field size, filled with ones. Keyword
        arguments are the same as that for the equivalent numpy function.

        Parameters
        ----------
        group : str
            Name of the group.

        See Also
        --------
        numpy.ones : See for a description of optional keywords.
        landlab.field.GraphFields.empty : Equivalent method that
            does not initialize the new array.
        landlab.field.GraphFields.zeros : Equivalent method that
            initializes the data to 0.

        Examples
        --------
        >>> from landlab.field import GraphFields
        >>> field = GraphFields()
        >>> field.new_field_location('node', 4)
        >>> field.ones('node')
        array([ 1.,  1.,  1.,  1.])
        >>> field.ones('node', dtype=int)
        array([1, 1, 1, 1])

        Note that a new field is *not* added to the collection of fields.

        >>> list(field.keys('node'))
        []

        LLCATS: FIELDCR
        """
        allocated = self.empty(*args, **kwds)
        allocated.fill(1)
        return allocated

    def zeros(self, *args, **kwds):
        """Array, initialized to 0, whose size is that of the field.

        Parameters
        ----------
        group : str
            Name of the group.

        Return a new array of the data field size, filled with zeros. Keyword
        arguments are the same as that for the equivalent numpy function.

        This method is not valid for the group *grid*.        
        
        See Also
        --------
        numpy.zeros : See for a description of optional keywords.
        landlab.field.GraphFields.empty : Equivalent method that does not
            initialize the new array.
        landlab.field.GraphFields.ones : Equivalent
            method that initializes the data to 1.

        Examples
        --------
        >>> from landlab.field import GraphFields
        >>> field = GraphFields()
        >>> field.new_field_location('node', 4)
        >>> field.zeros('node')
        array([ 0.,  0.,  0.,  0.])

        Note that a new field is *not* added to the collection of fields.

        >>> list(field.keys('node'))
        []

        LLCATS: FIELDCR
        """
        allocated = self.empty(*args, **kwds)
        allocated.fill(0)
        return allocated

    def add_field(self, *args, **kwds):
        """add_field(name, value_array, at='node', units='-', copy=False, noclobber=True)

        Add an array of values to the field.

        Add an array of data values to a collection of fields and associate it
        with the key, *name*. Use the *copy* keyword to, optionally, add a
        copy of the provided array.
        
        In the case of adding to the collection *grid*, the added field is a
        numpy scalar rather than a numpy array. 

        Parameters
        ----------
        name : str
            Name of the new field to add.
        value_array : numpy.array
            Array of values to add to the field.
        at : str, optional
            Grid location to store values. If not given, values are
            assumed to be on `node`.
        units : str, optional
            Optionally specify the units of the field.
        copy : boolean, optional
            If True, add a *copy* of the array to the field. Otherwise save add
            a reference to the array.
        noclobber : boolean, optional
            Raise an exception if adding to an already existing field.

        Returns
        -------
        numpy.array
            The data array added to the field. Depending on the *copy*
            keyword, this could be a copy of *value_array* or *value_array*
            itself.

        Raises
        ------
        ValueError :
            If *value_array* has a size different from the field.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab.field import GraphFields
        >>> field = GraphFields()
        >>> field.new_field_location('node', 4)
        >>> values = np.ones(4, dtype=int)
        >>> field.add_field('node', 'topographic__elevation', values)
        array([1, 1, 1, 1])

        A new field is added to the collection of fields. The saved value
        array is the same as the one initially created.

        >>> field.at_node['topographic__elevation'] is values
        True

        If you want to save a copy of the array, use the *copy* keyword. In
        addition, adding values to an existing field will remove the reference
        to the previously saved array. The *noclobber* keyword changes this
        behavior to raise an exception in such a case.

        >>> field.add_field('node', 'topographic__elevation', values,
        ...     copy=True, noclobber=False)
        array([1, 1, 1, 1])
        >>> field.at_node['topographic__elevation'] is values
        False
        >>> field.add_field('node', 'topographic__elevation', values,
        ...     noclobber=True) # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        FieldError: topographic__elevation

        LLCATS: FIELDCR
        """
        if len(args) == 3:
            at, name, value_array = args
        elif len(args) == 2:
            at, name, value_array = (kwds.pop('at', self._default_group),
                                     args[0], args[1])
        else:
            raise ValueError('number of arguments must be 2 or 3')

        units = kwds.get('units', '?')
        copy = kwds.get('copy', False)
        noclobber = kwds.get('noclobber', True)
        value_array = np.asarray(value_array)

        attrs = {'long_name': name}
        attrs['units'] = units

        if copy:
            value_array = value_array.copy()

        ds = getattr(self, 'at_' + at)

        if noclobber and name in ds:
            raise FieldError('{name}@{at}'.format(name=name, at=at))

        dims = (at, )
        if value_array.ndim > 1:
            dims += (name + '_per_' + at, )
            value_array = value_array.reshape((value_array.shape[0], -1))

        ds[name] = value_array
        return ds[name]

    def delete_field(self, loc, name):
        """Erases an existing field.

        Parameters
        ----------
        group : str
            Name of the group.
        name: str
            Name of the field.

        Raises
        ------
        KeyError
            If the named field does not exist.

        LLCATS: FIELDCR
        """
        try:
            ds = getattr(self, 'at_' + loc)
        except AttributeError:
            raise KeyError(loc)
        ds._ds = ds._ds.drop(name)

    def add_empty(self, *args, **kwds):
        """add_empty(name, at='node', units='-', noclobber=True)

        Create and add an uninitialized array of values to the field.

        Create a new array of the data field size, without initializing
        entries, and add it to the field as *name*. The *units* keyword gives
        the units of the new fields as a string. Remaining keyword arguments
        are the same as that for the equivalent numpy function.
        
        This method is not valid for the group *grid*.

        Parameters
        ----------
        name : str
            Name of the new field to add.
        at : str, optional
            Grid location to store values. If not given, values are
            assumed to be on `node`.
        units : str, optional
            Optionally specify the units of the field.
        noclobber : boolean, optional
            Raise an exception if adding to an already existing field.

        Returns
        -------
        array :
            A reference to the newly-created array.

        See Also
        --------
        numpy.empty : See for a description of optional keywords.
        landlab.field.GraphFields.empty : Equivalent method that
            does not initialize the new array.
        landlab.field.GraphFields.zeros : Equivalent method that
            initializes the data to 0.

        LLCATS: FIELDCR
        """
        if len(args) == 2:
            loc, name = args
        elif len(args) == 1:
            loc, name = kwds.pop('at'), args[0]
        else:
            raise ValueError('number of arguments must be 1 or 2')
        units = kwds.pop('units', '?')
        copy = kwds.pop('copy', False)
        noclobber = kwds.pop('noclobber', True)
        return self.add_field(name, self.empty(at=loc, **kwds), at=loc,
                              units=units, copy=copy, noclobber=noclobber)

    def add_ones(self, *args, **kwds):
        """add_ones(name, at='node', units='-', noclobber=True)

        Create and add an array of values, initialized to 1, to the field.

        Create a new array of the data field size, filled with ones, and
        add it to the field as *name*. The *units* keyword gives the units of
        the new fields as a string. Remaining keyword arguments are the same
        as that for the equivalent numpy function.
        
        This method is not valid for the group *grid*.

        Parameters
        ----------
        name : str
            Name of the new field to add.
        at : str, optional
            Grid location to store values. If not given, values are
            assumed to be on `node`.
        units : str, optional
            Optionally specify the units of the field.
        noclobber : boolean, optional
            Raise an exception if adding to an already existing field.

        Returns
        -------
        array :
            A reference to the newly-created array.

        See Also
        --------
        numpy.ones : See for a description of optional keywords.
        andlab.field.GraphFields.add_empty : Equivalent method that
            does not initialize the new array.
        andlab.field.GraphFields.add_zeros : Equivalent method that
            initializes the data to 0.

        Examples
        --------
        Add a new, named field to a collection of fields.

        >>> from landlab.field import GraphFields
        >>> field = GraphFields()
        >>> field.new_field_location('node', 4)
        >>> field.add_ones('node', 'topographic__elevation')
        array([ 1.,  1.,  1.,  1.])
        >>> list(field.keys('node'))
        ['topographic__elevation']
        >>> field['node']['topographic__elevation']
        array([ 1.,  1.,  1.,  1.])
        >>> field.at_node['topographic__elevation']
        array([ 1.,  1.,  1.,  1.])

        LLCATS: FIELDCR
        """
        data = self.add_empty(*args, **kwds)
        data.fill(1)
        return data

    def add_zeros(self, *args, **kwds):
        """add_zeros(name, at='node', units='-', noclobber=True)

        Create and add an array of values, initialized to 0, to the field.

        Create a new array of the data field size, filled with zeros, and
        add it to the field as *name*. The *units* keyword gives the units of
        the new fields as a string. Remaining keyword arguments are the same
        as that for the equivalent numpy function.

        Parameters
        ----------
        name : str
            Name of the new field to add.
        at : str, optional
            Grid location to store values. If not given, values are
            assumed to be on `node`.
        units : str, optional
            Optionally specify the units of the field.
        noclobber : boolean, optional
            Raise an exception if adding to an already existing field.

        Returns
        -------
        array :
            A reference to the newly-created array.

        See also
        --------
        numpy.zeros : See for a description of optional keywords.
        landlab.field.GraphFields.add_empty : Equivalent method that
            does not initialize the new array.
        landlab.field.GraphFields.add_ones : Equivalent method that
            initializes the data to 1.

        LLCATS: FIELDCR
        """
        data = self.add_empty(*args, **kwds)
        data.fill(0)
        return data

    def add_full(self, *args, **kwds):
        """Create and add an array of values, fill with `fill_value`.

        Parameters
        ----------
        name : str
            Name of the new field to add.
        fill_value : scalar
            Fill value.
        at : str, optional
            Grid location to store values. If not given, values are
            assumed to be on `node`.
        units : str, optional
            Optionally specify the units of the field.
        copy : boolean, optional
            If True, add a *copy* of the array to the field. Otherwise save add
            a reference to the array.
        noclobber : boolean, optional
            Raise an exception if adding to an already existing field.

        Returns
        -------
        array :
            A reference to the newly-created array.

        LLCATS: FIELDCR
        """
        if len(args) == 3:
            fill_value = args[2]
        elif len(args) == 2:
            fill_value = args[1]
        else:
            raise ValueError('number of arguments must be 2 or 3')

        data = self.add_empty(*args, **kwds)
        data.fill(fill_value)
        return data



import numpy as np
cimport numpy as np
cimport cython


DTYPE = np.int
ctypedef np.int_t DTYPE_INT_t
DTYPE_FLOAT = np.double
ctypedef np.double_t DTYPE_FLOAT_t


@cython.boundscheck(False)
def _calc_dists_to_channel(np.ndarray[np.uint8_t, ndim=1] ch_network,
                           np.ndarray[DTYPE_INT_t, ndim=1] flow_receivers,
                           np.ndarray[DTYPE_FLOAT_t, ndim=1] link_lengths,
                           np.ndarray[DTYPE_INT_t, ndim=1] stack_links,
                           np.ndarray[DTYPE_FLOAT_t, ndim=1] dist_to_ch,
                           DTYPE_INT_t num_nodes):
    """Calculate distance to nearest channel.

    Calculate the distances to the closest channel node for all nodes in the
    grid.

    Parameters
    ----------
    ch_network : node array
        integer logical map of which nodes contain channels.
    flow_receivers : node array
        ID of the next downstream node.
    link_lengths : num_d8_links-length array of floats
        The length of all links on the grid, including diagonals if present.
    stack_links : node array
        The ID of the link that leads to the downstream node.
    dists_to_ch : number_of_nodes-length array of floats
        The output array; the distance to the nearest channel node.
    num_nodes : int
        The number of nodes.
    """
    cdef DTYPE_INT_t node
    cdef DTYPE_INT_t node_iter
    cdef DTYPE_INT_t flag
    cdef DTYPE_FLOAT_t distance
    
    for node in range(num_nodes):
        if ch_network[node] == 1:
            # ^we're standing in a channel
            dist_to_ch[node] = 0.
        else:
            distance = 0
            flag = 0
            node_iter = node
            while flag == 0:  # as long as we haven't hit a channel yet...
                if flow_receivers[node_iter] == node_iter:
                    # ^if no flow receiver (boundary probably)
                    ch_network[node_iter] = 1
                    # ^convince the node it's a channel
                else:
                    distance += link_lengths[stack_links[node_iter]]
                    node_iter = flow_receivers[node_iter]
                if ch_network[node_iter] == 1:
                    # ^we've hit a channel
                    dist_to_ch[node] = distance
                    # ^save distance to channel
                    flag = 1



#! /usr/bin/env python
import numpy as np


from .nodestatus import (CLOSED_BOUNDARY, CORE_NODE, FIXED_GRADIENT_BOUNDARY,
                         FIXED_VALUE_BOUNDARY)
from .linkstatus import ACTIVE_LINK, INACTIVE_LINK, FIXED_LINK
from .linkstatus import (is_fixed_link, is_inactive_link, is_active_link,
                         set_status_at_link)
from ..utils.decorators import (cache_result_in_object,
                                make_return_array_immutable)
from .decorators import return_readonly_id_array


def create_nodes_at_diagonal(shape, out=None):
    """Create array of tail and head nodes for diagonals.

    Parameters
    ----------
    shape : tuple of *(n_rows, n_cols)*
        Shape as number of node rows and node columns.
    out : ndarray of shape *(n_diagonals, 2)*, optional
        Output buffer to place nodes at each diagonal.

    Returns
    -------
    out : ndarray of shape *(n_diagonals, 2)*
        Tail and head node for each diagonal.

    Examples
    --------
    >>> from landlab.grid.diagonals import create_nodes_at_diagonal
    >>> create_nodes_at_diagonal((3, 4)) # doctest: +NORMALIZE_WHITESPACE
    array([[ 0, 5], [ 1, 4], [ 1,  6], [ 2, 5], [ 2,  7], [ 3,  6],
           [ 4, 9], [ 5, 8], [ 5, 10], [ 6, 9], [ 6, 11], [ 7, 10]])
    """
    shape = np.asarray(shape)
    n_diagonals = np.prod(shape - 1) * 2
    n_nodes = np.prod(shape)
    if out is None:
        out = np.empty((n_diagonals, 2), dtype=int)

    nodes = np.arange(n_nodes).reshape(shape)

    out[::2, 0] = nodes[:-1, :-1].flat
    out[::2, 1] = nodes[1:, 1:].flat
    out[1::2, 0] = nodes[:-1, 1:].flat
    out[1::2, 1] = nodes[1:, :-1].flat

    return out


def create_diagonals_at_node(shape, out=None):
    """Create array of diagonals at node.

    Create an array that gives the diagonal touching each node
    for a structured grid of quadrilaterals. For each node, links
    are ordered clockwise from axis=1.

    Parameters
    ----------
    shape : tuple of *(n_rows, n_cols)*
        Shape as number of node rows and node columns.
    out : ndarray of shape *(n_nodes, 4)*, optional
        Output buffer to place diagonal ids at each node.

    Returns
    -------
    out : ndarray of shape *(n_nodes, 4)*
        Diagonals at node with -1 for missing diagonals.

    Examples
    --------
    >>> from landlab.grid.diagonals import create_diagonals_at_node
    >>> create_diagonals_at_node((3, 4))
    array([[ 0, -1, -1, -1],
           [ 2,  1, -1, -1],
           [ 4,  3, -1, -1],
           [-1,  5, -1, -1],
           [ 6, -1, -1,  1],
           [ 8,  7,  0,  3],
           [10,  9,  2,  5],
           [-1, 11,  4, -1],
           [-1, -1, -1,  7],
           [-1, -1,  6,  9],
           [-1, -1,  8, 11],
           [-1, -1, 10, -1]])
    """
    shape = np.asarray(shape)
    n_diagonals = np.prod(shape - 1) * 2
    n_nodes = np.prod(shape)
    if out is None:
        out = np.full((n_nodes, 4), -1, dtype=int)

    diagonals = np.full(shape + 1, -1, dtype=int)

    diagonals[1:-1, 1:-1] = np.arange(0, n_diagonals, 2).reshape(shape - 1)
    out[:, 0] = diagonals[1:, 1:].flat
    out[:, 2] = diagonals[:-1, :-1].flat

    diagonals[1:-1, 1:-1] = np.arange(1, n_diagonals, 2).reshape(shape - 1)
    out[:, 1] = diagonals[1:, :-1].flat
    out[:, 3] = diagonals[:-1, 1:].flat

    return out


def create_diagonal_dirs_at_node(shape, out=None):
    """Create array of diagonals directions at node.

    Parameters
    ----------
    shape : tuple of *(n_rows, n_cols)*
        Shape as number of node rows and node columns.
    out : ndarray of shape *(n_nodes, 4)*, optional
        Output buffer to place diagonal ids at each node.

    Returns
    -------
    out : ndarray of shape *(n_nodes, 4)*
        Diagonals at node with -1 for missing diagonals.

    Examples
    --------
    >>> from landlab.grid.diagonals import create_diagonals_at_node
    >>> create_diagonals_at_node((3, 4))
    array([[ 0, -1, -1, -1],
           [ 2,  1, -1, -1],
           [ 4,  3, -1, -1],
           [-1,  5, -1, -1],
           [ 6, -1, -1,  1],
           [ 8,  7,  0,  3],
           [10,  9,  2,  5],
           [-1, 11,  4, -1],
           [-1, -1, -1,  7],
           [-1, -1,  6,  9],
           [-1, -1,  8, 11],
           [-1, -1, 10, -1]])
    """
    shape = np.asarray(shape)
    n_diagonals = np.prod(shape - 1) * 2
    n_nodes = np.prod(shape)
    if out is None:
        out = np.full((n_nodes, 4), -1, dtype=int8)

    dirs = np.zeros(shape + 1, dtype=int8)

    dirs[1:-1, 1:-1] = np.arange(0, n_diagonals, 2).reshape(shape - 1)
    out[:, 0] = diagonals[1:, 1:].flat
    out[:, 2] = diagonals[:-1, :-1].flat

    diagonals[1:-1, 1:-1] = np.arange(1, n_diagonals, 2).reshape(shape - 1)
    out[:, 1] = diagonals[1:, :-1].flat
    out[:, 3] = diagonals[:-1, 1:].flat

    return out


class DiagonalsMixIn(object):

    """Add diagonals to a structured quad grid."""

    @property
    @cache_result_in_object()
    def number_of_diagonals(self):
        """Number of diagonals in the grid.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> grid = RasterModelGrid((4, 3))
        >>> grid.number_of_diagonals
        12
        """
        return 2 * np.prod(np.asarray(self.shape) - 1)

    @property
    @cache_result_in_object()
    @make_return_array_immutable
    def diagonals_at_node(self):
        """Diagonals attached to nodes.

        Returns
        -------
        ndarray of int, shape `(n_nodes, 4)`
            Diagonals at each node.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> grid = RasterModelGrid((4, 3))
        >>> grid.diagonals_at_node.shape == (grid.number_of_nodes, 4)
        True
        >>> grid.diagonals_at_node
        array([[ 0, -1, -1, -1], [ 2,  1, -1, -1], [-1,  3, -1, -1],
               [ 4, -1, -1,  1], [ 6,  5,  0,  3], [-1,  7,  2, -1],
               [ 8, -1, -1,  5], [10,  9,  4,  7], [-1, 11,  6, -1],
               [-1, -1, -1,  9], [-1, -1,  8, 11], [-1, -1, 10, -1]])

        LLCATS: NINF LINF CONN
        """
        return create_diagonals_at_node(self.shape)

    @property
    @cache_result_in_object()
    @make_return_array_immutable
    def diagonal_dirs_at_node(self):
        """Directions of diagonals attached to nodes.

        Array of diagonal directions relative to each node. If the
        diagonal is directed away from the node the direction is -1, if the
        diagonal is directed toward the node its direction is 1. Each
        node is assumed to have exactly four diagonals attached to it.
        However, perimeter nodes will have few diagonals (corner ndoes
        will only have one diagonal and edge nodes two). In such cases,
        placeholders of 0 are used.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> grid = RasterModelGrid((4, 3))
        >>> grid.diagonal_dirs_at_node.shape == (grid.number_of_nodes, 4)
        True
        >>> grid.diagonal_dirs_at_node
        array([[-1,  0,  0,  0], [-1, -1,  0,  0], [ 0, -1,  0,  0],
               [-1,  0,  0,  1], [-1, -1,  1,  1], [ 0, -1,  1,  0],
               [-1,  0,  0,  1], [-1, -1,  1,  1], [ 0, -1,  1,  0],
               [ 0,  0,  0,  1], [ 0,  0,  1,  1], [ 0,  0,  1,  0]],
               dtype=int8)

        The lower-right corner node only has one diagonal.

        >>> grid.diagonal_dirs_at_node[2]
        array([ 0, -1,  0,  0], dtype=int8)

        A node on one of the edges has two diagonals.

        >>> grid.diagonal_dirs_at_node[3]
        array([-1,  0,  0,  1], dtype=int8)
        """
        diagonals_at_node = self.diagonals_at_node
        dirs_at_node = np.zeros_like(diagonals_at_node, dtype=np.int8)
        dirs_at_node[diagonals_at_node >= 0] = 1
        dirs_at_node[:, (0, 1)] *= -1

        return dirs_at_node

    @property
    @cache_result_in_object()
    @make_return_array_immutable
    def diagonal_adjacent_nodes_at_node(self):
        """Get adjacent nodes along diagonals.

        Order is the landlab standard, counterclockwise starting from
        east.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> grid = RasterModelGrid((4, 3))
        >>> grid.diagonal_adjacent_nodes_at_node
        array([[ 4, -1, -1, -1], [ 5,  3, -1, -1], [-1,  4, -1, -1],
               [ 7, -1, -1,  1], [ 8,  6,  0,  2], [-1,  7,  1, -1],
               [10, -1, -1,  4], [11,  9,  3,  5], [-1, 10,  4, -1],
               [-1, -1, -1,  7], [-1, -1,  6,  8], [-1, -1,  7, -1]])

        LLCATS: NINF CONN
        """
        node_is_at_tail = np.choose(self.diagonal_dirs_at_node + 1,
                                    np.array((1, -1, 0), dtype=np.int8))
        out = self.nodes_at_diagonal[self.diagonals_at_node, node_is_at_tail]
        out[node_is_at_tail == -1] = -1

        return out

    @property
    @cache_result_in_object()
    @make_return_array_immutable
    def d8_adjacent_nodes_at_node(self):
        return np.vstack((super(DiagonalsMixIn, self).adjacent_nodes_at_node,
                          self.diagonal_adjacent_nodes_at_node))

    @property
    @cache_result_in_object()
    @make_return_array_immutable
    def nodes_at_diagonal(self):
        """Nodes at diagonal tail and head.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> grid = RasterModelGrid((4, 3))
        >>> grid.nodes_at_diagonal.shape == (grid.number_of_diagonals, 2)
        True
        >>> grid.nodes_at_diagonal
        array([[ 0,  4], [ 1,  3], [ 1,  5], [ 2,  4],
               [ 3,  7], [ 4,  6], [ 4,  8], [ 5,  7],
               [ 6, 10], [ 7,  9], [ 7, 11], [ 8, 10]])

        The first column is diagonal tail and the second the head.

        >>> grid.diagonals_at_node[3]
        array([ 4, -1, -1,  1])
        >>> grid.nodes_at_diagonal[(4, 1), ]
        array([[3, 7],
               [1, 3]])
        >>> grid.diagonal_dirs_at_node[3]
        array([-1,  0,  0,  1], dtype=int8)
        """
        return create_nodes_at_diagonal(self.shape)

    @property
    @cache_result_in_object()
    def number_of_d8(self):
        """Number of links and diagonals.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> grid = RasterModelGrid((4, 3))
        >>> grid.number_of_d8
        29
        """
        return (super(DiagonalsMixIn, self).number_of_links +
                self.number_of_diagonals)

    @property
    @cache_result_in_object()
    @make_return_array_immutable
    def nodes_at_d8(self):
        return np.vstack((self.nodes_at_link, self.nodes_at_diagonal))

    @property
    @cache_result_in_object()
    @make_return_array_immutable
    def d8s_at_node(self):
        """Links and diagonals attached to nodes.

        Returns
        -------
        ndarray of int, shape `(n_nodes, 8)`
            Links and diagonals at each node.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid

        >>> grid = RasterModelGrid((3, 4))
        >>> grid.d8s_at_node.shape == (grid.number_of_nodes, 8)
        True
        >>> grid.d8s_at_node
        array([[ 0,  3, -1, -1, 17, -1, -1, -1],
               [ 1,  4,  0, -1, 19, 18, -1, -1],
               [ 2,  5,  1, -1, 21, 20, -1, -1],
               [-1,  6,  2, -1, -1, 22, -1, -1],
               [ 7, 10, -1,  3, 23, -1, -1, 18],
               [ 8, 11,  7,  4, 25, 24, 17, 20],
               [ 9, 12,  8,  5, 27, 26, 19, 22],
               [-1, 13,  9,  6, -1, 28, 21, -1],
               [14, -1, -1, 10, -1, -1, -1, 24],
               [15, -1, 14, 11, -1, -1, 23, 26],
               [16, -1, 15, 12, -1, -1, 25, 28],
               [-1, -1, 16, 13, -1, -1, 27, -1]])
        >>> np.all(grid.d8s_at_node[:, :4] == grid.links_at_node)
        True

        >>> diagonals_at_node = grid.d8s_at_node[:, 4:] - grid.number_of_links
        >>> diagonals_at_node[grid.d8s_at_node[:, 4:] == -1] = -1
        >>> np.all(diagonals_at_node == grid.diagonals_at_node)
        True

        LLCATS: NINF LINF CONN
        """
        diagonals_at_node = self.diagonals_at_node.copy()
        diagonals_at_node[diagonals_at_node >= 0] += self.number_of_links
        return np.hstack((super(DiagonalsMixIn, self).links_at_node,
                          diagonals_at_node))
                          # self.diagonals_at_node + self.number_of_links))

    @property
    @cache_result_in_object()
    @make_return_array_immutable
    def d8_dirs_at_node(self):
        return np.hstack((super(DiagonalsMixIn, self).link_dirs_at_node,
                          self.diagonal_dirs_at_node))

    @property
    # @cache_result_in_object()
    @make_return_array_immutable
    def d8_status_at_node(self):
        return self.status_at_d8[self.d8s_at_node]

    @property
    @cache_result_in_object()
    @make_return_array_immutable
    def length_of_diagonal(self):
        return np.sqrt(
            np.power(np.diff(self.xy_of_node[self.nodes_at_diagonal], axis=1),
                     2.).sum(axis=2)).flatten()

    @property
    @cache_result_in_object()
    @make_return_array_immutable
    def length_of_d8(self):
        """Length of links and diagonals.

        Return the lengths links and diagonals in the grid. Links are
        listed first and then diagonals.

        Returns
        -------
        ndarray of float
            Link lengths.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> grid = RasterModelGrid((3, 3), spacing=(3, 4))

        >>> grid.length_of_link
        array([ 4.,  4.,  3.,  3.,  3.,  4.,  4.,  3.,  3.,  3.,  4.,  4.])

        >>> grid.length_of_d8 # doctest: +NORMALIZE_WHITESPACE
        array([ 4.,  4.,  3.,  3.,  3.,
                4.,  4.,  3.,  3.,  3.,
                4.,  4.,  5.,  5.,  5.,
                5.,  5.,  5.,  5.,  5.])

        LLCATS: LINF MEAS
        """
        return np.hstack((super(DiagonalsMixIn, self).length_of_link,
                          self.length_of_diagonal))

    def reset_status_at_node(self):
        super(DiagonalsMixIn, self).reset_status_at_node()
        attrs = ['_status_at_diagonal', '_diagonal_status_at_node',
                 '_active_diagonals', '_active_diagonal_dirs_at_node',
                 '_status_at_d8', '_active_d8', '_active_d8_dirs_at_node']

        for attr in attrs:
            try:
                del self.__dict__[attr]
            except KeyError:
                pass

    @property
    @cache_result_in_object()
    @make_return_array_immutable
    def status_at_diagonal(self):
        """Status at diagonals.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> grid = RasterModelGrid((4, 3))

        An inactive link (or diagonal) is one that joins two
        boundary nodes or has one end that is a closed boundary.

        >>> grid.status_at_node
        array([1, 1, 1,
               1, 0, 1,
               1, 0, 1,
               1, 1, 1], dtype=uint8)
        >>> grid.BC_NODE_IS_CORE, grid.BC_NODE_IS_FIXED_VALUE
        (0, 1)

        Diagonals that connect two boundary nodes are inactive.

        >>> grid.status_at_diagonal
        array([0, 4, 4, 0,
               0, 0, 0, 0,
               4, 0, 0, 4], dtype=uint8)
        >>> grid.BC_LINK_IS_ACTIVE, grid.BC_LINK_IS_INACTIVE
        (0, 4)

        By setting a node to closed that wasn't before, a new link
        become inactive.

        >>> grid.status_at_node[5] = grid.BC_NODE_IS_CLOSED
        >>> grid.status_at_diagonal
        array([0, 4, 4, 0,
               0, 0, 0, 4,
               4, 0, 0, 4], dtype=uint8)
        """
        return set_status_at_link(self.status_at_node[self.nodes_at_diagonal])

    @property
    @cache_result_in_object()
    @make_return_array_immutable
    def diagonal_status_at_node(self):
        return self.status_at_diagonal[self.diagonals_at_node]

    @property
    @cache_result_in_object()
    @return_readonly_id_array
    def active_diagonals(self):
        return np.where(self.status_at_diagonal == ACTIVE_LINK)[0]

    @property
    @cache_result_in_object()
    @make_return_array_immutable
    def active_diagonal_dirs_at_node(self):
        return np.choose(self.diagonal_status_at_node == ACTIVE_LINK,
                         (0, self.diagonal_dirs_at_node))

    @property
    @cache_result_in_object()
    @make_return_array_immutable
    def status_at_d8(self):
        return np.hstack((super(DiagonalsMixIn, self).status_at_link,
                          self.status_at_diagonal))

    @property
    @cache_result_in_object()
    @return_readonly_id_array
    def active_d8(self):
        return np.where(self.status_at_d8 == ACTIVE_LINK)[0]

    @property
    @cache_result_in_object()
    @make_return_array_immutable
    def active_d8_dirs_at_node(self):
        return np.choose(self.d8_status_at_node == ACTIVE_LINK,
                         (0, self.d8_dirs_at_node))



"""
Unit tests for landlab.components.pet.potential_evapotranspiration_field
"""
import pytest
from numpy.testing import assert_array_almost_equal
import numpy as np


(_SHAPE, _SPACING, _ORIGIN) = ((20, 20), (10e0, 10e0), (0., 0.))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def test_name(pet):
    assert pet.name == 'Potential Evapotranspiration'


def test_input_var_names(pet):
    assert pet.input_var_names == ('radiation__ratio_to_flat_surface', )


def test_output_var_names(pet):
    assert sorted(pet.output_var_names) == [
        'radiation__incoming_shortwave_flux',
        'radiation__net_flux',
        'radiation__net_longwave_flux',
        'radiation__net_shortwave_flux',
        'surface__potential_evapotranspiration_rate',
    ]


def test_var_units(pet):
    assert set(pet.input_var_names) | set(pet.output_var_names) == set(dict(pet.units).keys())

    assert pet.var_units('radiation__incoming_shortwave_flux') == 'W/m^2'
    assert pet.var_units('radiation__net_flux') == 'W/m^2'
    assert pet.var_units('radiation__net_longwave_flux') == 'W/m^2'
    assert pet.var_units('radiation__net_shortwave_flux') == 'W/m^2'
    assert pet.var_units('radiation__ratio_to_flat_surface') == 'None'
    assert pet.var_units('surface__potential_evapotranspiration_rate') == 'mm'


def test_grid_shape(pet):
    assert pet.grid.number_of_node_rows == _SHAPE[0]
    assert pet.grid.number_of_node_columns == _SHAPE[1]


def test_grid_x_extent(pet):
    assert pet.grid.extent[1] == (_SHAPE[1] - 1) * _SPACING[1]


def test_grid_y_extent(pet):
    assert pet.grid.extent[0] == (_SHAPE[0] - 1) * _SPACING[0]


def test_field_getters(pet):
    for name in pet.grid['node']:
        field = pet.grid['node'][name]
        assert isinstance(field, np.ndarray)
        assert field.shape == (pet.grid.number_of_node_rows *
                               pet.grid.number_of_node_columns, )
                      
    for name in pet.grid['cell']:
        field = pet.grid['cell'][name]
        assert isinstance(field, np.ndarray)
        assert field.shape == (pet.grid.number_of_cell_rows *
                               pet.grid.number_of_cell_columns, )

    with pytest.raises(KeyError):
        pet.grid['not_a_var_name']


def test_field_initialized_to_zero(pet):
    for name in pet.grid['node']:
        field = pet.grid['node'][name]
        assert_array_almost_equal(field, np.zeros(pet.grid.number_of_nodes))
    for name in pet.grid['cell']:
        field = pet.grid['cell'][name]
        assert_array_almost_equal(field, np.zeros(pet.grid.number_of_cells))



import numpy as np
from landlab.components.erosion_deposition.generalized_erosion_deposition import (_GeneralizedErosionDeposition,
                                                            DEFAULT_MINIMUM_TIME_STEP)
from landlab.utils.return_array import return_array_at_node
from .cfuncs import calculate_qs_in

ROOT2 = np.sqrt(2.0)    # syntactic sugar for precalculated square root of 2
TIME_STEP_FACTOR = 0.5  # factor used in simple subdivision solver
DEFAULT_MINIMUM_TIME_STEP = 0.001  # default minimum time step duration

class ErosionDeposition(_GeneralizedErosionDeposition):
    """
    Erosion-Deposition model in the style of Davy and Lague (2009)

    Component written by C. Shobe, K. Barnhart, and G. Tucker.
    """

    _name= 'ErosionDeposition'

    _input_var_names = (
        'flow__receiver_node',
        'flow__upstream_node_order',
        'topographic__steepest_slope',
        'drainage_area',
    )

    _output_var_names = (
        'topographic__elevation'
    )

    _var_units = {
        'flow__receiver_node': '-',
        'flow__upstream_node_order': '-',
        'topographic__steepest_slope': '-',
        'drainage_area': 'm**2',
        'topographic__elevation': 'm',
    }

    _var_mapping = {
        'flow__receiver_node': 'node',
        'flow__upstream_node_order': 'node',
        'topographic__steepest_slope': 'node',
        'drainage_area': 'node',
        'topographic__elevation': 'node',
    }

    _var_doc = {
        'flow__receiver_node':
            'Node array of receivers (node that receives flow from current '
            'node)',
        'flow__upstream_node_order':
            'Node array containing downstream-to-upstream ordered list of '
            'node IDs',
        'topographic__steepest_slope':
            'Topographic slope at each node',
        'drainage_area':
            "Upstream accumulated surface area contributing to the node's "
            "discharge",
        'topographic__elevation':
            'Land surface topographic elevation',
    }

    def __init__(self, grid, K=None, phi=None, v_s=None,
                 m_sp=None, n_sp=None, sp_crit=0.0, F_f=0.0,
                 discharge_field='surface_water__discharge',
                 solver='basic',
                 dt_min=DEFAULT_MINIMUM_TIME_STEP,
                 **kwds):
        """Initialize the ErosionDeposition model.

        Parameters
        ----------
        grid : ModelGrid
            Landlab ModelGrid object
        K : float, field name, or array
            Erodibility for substrate (units vary).
        phi : float
            Sediment porosity [-].
        v_s : float
            Effective settling velocity for chosen grain size metric [L/T].
        m_sp : float
            Drainage area exponent (units vary)
        n_sp : float
            Slope exponent (units vary)
        sp_crit : float, field name, or array
            Critical stream power to erode substrate [E/(TL^2)]
        F_f : float
            Fraction of eroded material that turns into "fines" that do not
            contribute to (coarse) sediment load. Defaults to zero.
        discharge_field : float, field name, or array
            Discharge [L^2/T].
        solver : string
            Solver to use. Options at present include:
                (1) 'basic' (default): explicit forward-time extrapolation.
                    Simple but will become unstable if time step is too large.
                (2) 'adaptive': adaptive time-step solver that estimates a
                    stable step size based on the shortest time to "flattening"
                    among all upstream-downstream node pairs.

        Examples
        ---------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.components.flow_routing import FlowRouter
        >>> from landlab.components import DepressionFinderAndRouter
        >>> from landlab.components import ErosionDeposition
        >>> from landlab.components import FastscapeEroder
        >>> np.random.seed(seed = 5000)

        Define grid and initial topography:
            -5x5 grid with baselevel in the lower left corner
            -all other boundary nodes closed
            -Initial topography is plane tilted up to the upper right + noise

        >>> nr = 5
        >>> nc = 5
        >>> dx = 10
        >>> mg = RasterModelGrid((nr, nc), 10.0)
        >>> _ = mg.add_zeros('node', 'topographic__elevation')
        >>> mg['node']['topographic__elevation'] += (mg.node_y/10 +
        ...        mg.node_x/10 + np.random.rand(len(mg.node_y)) / 10)
        >>> mg.set_closed_boundaries_at_grid_edges(bottom_is_closed=True,
        ...                                               left_is_closed=True,
        ...                                               right_is_closed=True,
        ...                                               top_is_closed=True)
        >>> mg.set_watershed_boundary_condition_outlet_id(0,\
                mg['node']['topographic__elevation'], -9999.)
        >>> fsc_dt = 100.
        >>> ed_dt = 1.

        Check initial topography

        >>> mg.at_node['topographic__elevation'] # doctest: +NORMALIZE_WHITESPACE
        array([ 0.02290479,  1.03606698,  2.0727653 ,  3.01126678,  4.06077707,
            1.08157495,  2.09812694,  3.00637448,  4.07999597,  5.00969486,
            2.04008677,  3.06621577,  4.09655859,  5.04809001,  6.02641123,
            3.05874171,  4.00585786,  5.0595697 ,  6.04425233,  7.05334077,
            4.05922478,  5.0409473 ,  6.07035008,  7.0038935 ,  8.01034357])

        Instantiate Fastscape eroder, flow router, and depression finder

        >>> fsc = FastscapeEroder(mg, K_sp=.001, m_sp=.5, n_sp=1)
        >>> fr = FlowRouter(mg) #instantiate
        >>> df = DepressionFinderAndRouter(mg)

        Burn in an initial drainage network using the Fastscape eroder:

        >>> for x in range(100):
        ...     fr.run_one_step()
        ...     df.map_depressions()
        ...     flooded = np.where(df.flood_status==3)[0]
        ...     fsc.run_one_step(dt = fsc_dt, flooded_nodes=flooded)
        ...     mg.at_node['topographic__elevation'][0] -= 0.001 #uplift

        Instantiate the E/D component:

        >>> ed = ErosionDeposition(mg, K=0.00001, phi=0.0, v_s=0.001,
        ...                        m_sp=0.5, n_sp = 1.0, sp_crit=0)

        Now run the E/D component for 2000 short timesteps:

        >>> for x in range(2000): #E/D component loop
        ...     fr.run_one_step()
        ...     df.map_depressions()
        ...     flooded = np.where(df.flood_status==3)[0]
        ...     ed.run_one_step(dt = ed_dt, flooded_nodes=flooded)
        ...     mg.at_node['topographic__elevation'][0] -= 2e-4 * ed_dt

        Now we test to see if topography is right:

        >>> np.around(mg.at_node['topographic__elevation'], decimals=3) # doctest: +NORMALIZE_WHITESPACE
        array([-0.477,  1.036,  2.073,  3.011,  4.061,  1.082, -0.08 , -0.065,
           -0.054,  5.01 ,  2.04 , -0.065, -0.065, -0.053,  6.026,  3.059,
           -0.054, -0.053, -0.035,  7.053,  4.059,  5.041,  6.07 ,  7.004,
            8.01 ])
        """
        super(ErosionDeposition, self).__init__(grid, m_sp=m_sp, n_sp=n_sp,
                                                phi=phi, F_f=F_f, v_s=v_s,
                                                dt_min=dt_min,
                                                discharge_field=discharge_field)


        self._grid = grid #store grid

        # E/D specific inits.

        # K's and critical values can be floats, grid fields, or arrays
        self.K = return_array_at_node(grid, K)
        self.sp_crit = return_array_at_node(grid, sp_crit)

        # Handle option for solver
        if solver == 'basic':
            self.run_one_step = self.run_one_step_basic
        elif solver == 'adaptive':
            self.run_one_step = self.run_with_adaptive_time_step_solver
            self.time_to_flat = np.zeros(grid.number_of_nodes)
        else:
            raise ValueError("Parameter 'solver' must be one of: "
                             + "'basic', 'adaptive'")

    def _calc_erosion_rates(self):
        """Calculate erosion rates"""
        omega = self.K * self.Q_to_the_m * np.power(self.slope, self.n_sp)
        omega_over_sp_crit = np.divide(omega, self.sp_crit,
                             out=np.zeros_like(omega), where=self.sp_crit!=0)

        self.erosion_term = omega - self.sp_crit * (1.0 - np.exp(-omega_over_sp_crit))

    def run_one_step_basic(self, dt=1.0, flooded_nodes=[], **kwds):
        """Calculate change in rock and alluvium thickness for
           a time period 'dt'.

        Parameters
        ----------
        dt : float
            Model timestep [T]
        flooded_nodes : array
            Indices of flooded nodes, passed from flow router
        """

        self._calc_hydrology()
        self._calc_erosion_rates()

        self.erosion_term[flooded_nodes] = 0.0
        self.qs_in[:] = 0.0

        #iterate top to bottom through the stack, calculate qs
        # cythonized version of calculating qs_in
        calculate_qs_in(np.flipud(self.stack),
                        self.flow_receivers,
                        self.cell_area_at_node,
                        self.q,
                        self.qs,
                        self.qs_in,
                        self.erosion_term,
                        self.v_s,
                        self.F_f,
                        self.phi)

        self.depo_rate[:] = 0.0
        self.depo_rate[self.q > 0] = (self.qs[self.q > 0] * \
                                         (self.v_s / self.q[self.q > 0]))

        #topo elev is old elev + deposition - erosion
        cores = self.grid.core_nodes
        self.topographic__elevation[cores] += (((self.depo_rate[cores] /
                            (1 - self.phi))
                              - self.erosion_term[cores]) * dt)

    def run_with_adaptive_time_step_solver(self, dt=1.0, flooded_nodes=[],
                                           **kwds):
        """CHILD-like solver that adjusts time steps to prevent slope
        flattening."""

        # Initialize remaining_time, which records how much of the global time
        # step we have yet to use up.
        remaining_time = dt

        z = self._grid.at_node['topographic__elevation']
        r = self.flow_receivers
        dzdt = np.zeros(len(z))
        cores = self._grid.core_nodes

        first_iteration = True

        # Outer WHILE loop: keep going until time is used up
        while remaining_time > 0.0:

            # Update all the flow-link slopes.
            #
            # For the first iteration, we assume this has already been done
            # outside the component (e.g., by flow router), but we need to do
            # it ourselves on subsequent iterations.
            if not first_iteration:
                # update the link slopes
                self._update_flow_link_slopes()
                # update where nodes are flooded. This shouuldn't happen because
                # of the dynamic timestepper, but just incase, we update here.
                new_flooded_nodes = np.where(self.slope<0)[0]
                flooded_nodes = np.asarray(np.unique(np.concatenate((flooded_nodes,
                                                          new_flooded_nodes))), dtype=np.int64)
            else:
                first_iteration = False

            # Calculate rates of entrainment
            self._calc_hydrology()
            self._calc_erosion_rates()
            self.erosion_term[flooded_nodes] = 0.0
            self.qs_in[:] = 0.0

            # Sweep through nodes from upstream to downstream, calculating Qs.
            calculate_qs_in(np.flipud(self.stack),
                            self.flow_receivers,
                            self.cell_area_at_node,
                            self.q,
                            self.qs,
                            self.qs_in,
                            self.erosion_term,
                            self.v_s,
                            self.F_f,
                            self.phi)

            # Use Qs to calculate deposition rate at each node.
            self.depo_rate[:] = 0.0
            self.depo_rate[self.q > 0] = (self.qs[self.q > 0]
                                          * (self.v_s / self.q[self.q > 0]))

            # Rate of change of elevation at core nodes:
            dzdt[cores] = (self.depo_rate[cores] / (1 - self.phi)) - self.erosion_term[cores]

            # Difference in elevation between each upstream-downstream pair
            zdif = z - z[r]

            # Rate of change of the *difference* in elevation between each
            # upstream-downstream pair.
            rocdif = dzdt - dzdt[r]

            # (Re)-initialize the array that will contain "time to (almost)
            # flat" for each node (relative to its downstream neighbor).
            self.time_to_flat[:] = remaining_time

            # Find locations where the upstream and downstream node elevations
            # are converging (e.g., the upstream one is eroding faster than its
            # downstream neighbor)
            converging = np.where(rocdif < 0.0)[0]

            # Find the time to (almost) flat by dividing difference by rate of
            # change of difference, and then multiplying by a "safety factor"
            self.time_to_flat[converging] = - (TIME_STEP_FACTOR
                                               * zdif[converging]
                                              / rocdif[converging])

            # Mask out pairs where the source at the same or lower elevation
            # as its downstream neighbor (e.g., because it's a pit or a lake).
            # Here, masking out means simply assigning the remaining time in
            # the global time step.
            self.time_to_flat[np.where(zdif <= 0.0)[0]] = remaining_time
            self.time_to_flat[flooded_nodes] = remaining_time

            # From this, find the maximum stable time step. If it is smaller
            # than our tolerance, report and quit.
            dt_max = np.amin(self.time_to_flat)
            if dt_max < self.dt_min:
                dt_max = self.dt_min

            # Finally, apply dzdt to all nodes for a (sub)step of duration
            # dt_max
            z[cores] += dzdt[cores] * dt_max

            # Update remaining time and continue the loop
            remaining_time -= dt_max



import numpy as np

from ..voronoi.voronoi import VoronoiGraph
from ...utils.decorators import store_result_in_grid


def number_of_nodes(shape, node_layout='rect'):
    """Get the number of nodes in a hex graph.

    Parameters
    ----------
    shape : tuple of int
        Number of rows and columns of the hex grid. The first value
        is the number of nodes in the first column and the second the
        number of nodes in the first column.

    Examples
    --------
    >>> from landlab.graph.hex.hex import number_of_nodes
    >>> number_of_nodes((3, 2))
    6
    >>> number_of_nodes((2, 4))
    8
    >>> number_of_nodes((4, 2))
    8

    >>> number_of_nodes((3, 2), node_layout='hex')
    7
    >>> number_of_nodes((2, 4), node_layout='hex')
    9
    >>> number_of_nodes((4, 2), node_layout='hex')
    12

    >>> number_of_nodes((3, 2), node_layout='rect1')
    7
    >>> number_of_nodes((2, 4), node_layout='rect1')
    9
    >>> number_of_nodes((4, 2), node_layout='rect1')
    10
    """
    if node_layout not in ('rect', 'hex', 'rect1'):
        raise ValueError('node_layout not understood')

    n_rows, n_cols = shape

    if node_layout == 'rect':
        return n_rows * n_cols
    elif node_layout == 'hex':
        return  n_rows * n_cols + (n_rows // 2) ** 2
    elif node_layout == 'rect1':
        return (2 * n_cols + 1) * (n_rows // 2) + n_cols * (n_rows % 2)


def setup_perimeter_nodes(shape, orientation='horizontal', node_layout='rect'):
    from .ext.hex import fill_perimeter_nodes, fill_hex_perimeter_nodes

    n_perimeter_nodes = 2 * shape[0] + 2 * (shape[1] - 2)
    if node_layout in ('hex', 'rect1'):
        n_perimeter_nodes += (shape[0] + 1) % 2
    perimeter_nodes = np.empty(n_perimeter_nodes, dtype=int)

    if node_layout == 'hex':
        fill_hex_perimeter_nodes(shape, perimeter_nodes)
    else:
        fill_perimeter_nodes(shape, perimeter_nodes)

    return perimeter_nodes


def setup_xy_of_node(shape, spacing=1., origin=(0., 0.),
                     orientation='horizontal', node_layout='rect'):
    """Create arrays of coordinates of a node on a hex grid.

    Parameters
    ----------
    shape : tuple of int
        Number of rows and columns of the hex grid. The first value
        is the number of nodes in the first column and the second the
        number of nodes in the first column.
    spacing : float, optional
        Length of links.
    origin : tuple of float, optional
        Coordinates of lower-left corner of the grid.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.graph.hex.hex import setup_xy_of_node

    >>> x, y = setup_xy_of_node((3, 2))
    >>> x
    array([ 0. ,  1. ,  0.5,  1.5,  0. ,  1. ])
    >>> y / (np.sqrt(3) / 2.)
    array([ 0.,  0.,  1.,  1.,  2.,  2.])
    >>> x, y = setup_xy_of_node((2, 2))
    >>> x
    array([ 0. ,  1. ,  0.5,  1.5])
    >>> y / (np.sqrt(3) / 2.)
    array([ 0.,  0.,  1.,  1.])

    >>> x, y = setup_xy_of_node((2, 2), spacing=2, origin=(1, 2))
    >>> x
    array([ 2.,  4.,  3.,  5.])
    >>> (y - 1) / (np.sqrt(3) / 2.)
    array([ 0.,  0.,  2.,  2.])
    """
    from .ext.hex import get_xy_of_node, fill_xy_of_node, fill_hex_xy_of_node

    if orientation == 'vertical':
        return setup_xy_of_node((shape[1], shape[0]), spacing=spacing,
                                origin=(origin[1], origin[0]),
                                node_layout=node_layout,
                                orientation='horizontal')[::-1]

    n_nodes = number_of_nodes(shape, node_layout=node_layout)

    x_of_node = np.empty((n_nodes, ), dtype=float)
    y_of_node = np.empty((n_nodes, ), dtype=float)

    if node_layout == 'rect':
        fill_xy_of_node(shape, x_of_node, y_of_node)
    elif node_layout == 'hex':
        fill_hex_xy_of_node(shape, x_of_node, y_of_node)
    elif node_layout == 'rect1':
        get_xy_of_node(shape, x_of_node, y_of_node)

    x_of_node *= spacing
    x_of_node += origin[1]
    y_of_node *= (spacing * np.sin(np.pi / 3.))
    y_of_node += origin[0]

    return (x_of_node, y_of_node)


class HexGraph(VoronoiGraph):

    """Graph of a structured grid of triangles.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.graph import HexGraph

    >>> graph = HexGraph((3, 2))
    >>> graph.number_of_nodes == 6
    True
    >>> np.round(graph.y_of_node * 2. / np.sqrt(3))
    ...     # doctest: +NORMALIZE_WHITESPACE
    array([ 0.,  0.,  1.,  1.,  2.,  2.])
    >>> graph.x_of_node # doctest: +NORMALIZE_WHITESPACE
    array([ 0. ,  1. ,  0.5,  1.5,  0. ,  1. ])
    """

    def __init__(self, shape, spacing=1., origin=(0., 0.),
                 orientation='horizontal', node_layout='rect'):
        """Create a structured grid of triangles.

        Parameters
        ----------
        shape : tuple of int
            Number of rows and columns of the hex grid. The first value
            is the number of nodes in the first column and the second the
            number of nodes in the first column.
        spacing : float, optional
            Length of links.
        origin : tuple of float, optional
            Coordinates of lower-left corner of the grid.
        orientation: {'horizontal', 'vertical'}
            Specify if triangles should be laid out in rows or columns.
        node_layout: {'rect', 'hex', 'rect1'}
            Specify the overall layout of the nodes. Use *rect* for
            the layout to approximate a rectangle and *hex* for
            a hexagon.
        """
        try:
            spacing = float(spacing)
        except TypeError:
            raise TypeError('spacing must be a float')

        self._shape = tuple(shape)

        if node_layout not in ('rect', 'hex', 'rect1'):
            raise ValueError('node_layout not understood')
        else:
            self._node_layout = node_layout

        if orientation not in ('horizontal', 'vertical'):
            raise ValueError('orientation not understood')
        else:
            self._orientation = orientation

        x_of_node, y_of_node = setup_xy_of_node(shape, spacing=spacing,
                                                origin=origin,
                                                orientation=orientation,
                                                node_layout=node_layout)
        if node_layout == 'hex':
            max_node_spacing = shape[1] + shape[0] / 2 + 2
            max_node_spacing = None
        elif node_layout == 'rect':
            max_node_spacing = shape[1] + 1
        elif node_layout == 'rect1':
            max_node_spacing = shape[1] + 1

        VoronoiGraph.__init__(self,
            (y_of_node, x_of_node), xy_sort=True, rot_sort=True,
            max_node_spacing=max_node_spacing)

    @property
    def shape(self):
        return self._shape

    @property
    def orientation(self):
        return self._orientation

    @property
    def node_layout(self):
        return self._node_layout

    @property
    # @store_result_in_grid()
    def perimeter_nodes(self):
        return setup_perimeter_nodes(self.shape, self.orientation,
                                     self.node_layout)



import pytest

from landlab import RasterModelGrid
from landlab.components.plant_competition_ca.plant_competition_ca import VegCA


@pytest.fixture
def ca_veg():
    grid = RasterModelGrid((20, 20), spacing=10e0)
    return VegCA(grid)




from landlab import Component
from ...utils.decorators import use_file_name_or_kwds
import numpy as np


_VALID_METHODS = set(['Grid'])


def _assert_method_is_valid(method):
    if method not in _VALID_METHODS:
        raise ValueError('%s: Invalid method name' % method)


class Radiation(Component):

    """Compute 1D and 2D total incident shortwave radiation.

    Landlab component that computes 1D and 2D total incident shortwave
    radiation. This code also computes relative incidence shortwave radiation
    compared to a flat surface. Ref: Bras, Rafael L. Hydrology: an
    introduction to hydrologic science. Addison Wesley Publishing Company,
    1990.

    .. codeauthor:: Sai Nudurupati & Erkan Istanbulluoglu

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import Radiation
    >>> import numpy as np

    >>> grid = RasterModelGrid((5, 4), spacing=(0.2, 0.2))
    >>> rad = Radiation(grid)
    >>> rad.name
    'Radiation'
    >>> rad.input_var_names
    ('topographic__elevation',)
    >>> sorted(rad.output_var_names) # doctest: +NORMALIZE_WHITESPACE
    ['radiation__incoming_shortwave_flux',
     'radiation__net_shortwave_flux',
     'radiation__ratio_to_flat_surface']
    >>> sorted(rad.units) # doctest: +NORMALIZE_WHITESPACE
    [('radiation__incoming_shortwave_flux', 'W/m^2'),
     ('radiation__net_shortwave_flux', 'W/m^2'),
     ('radiation__ratio_to_flat_surface', 'None'),
     ('topographic__elevation', 'm')]

    >>> rad.grid.number_of_node_rows
    5
    >>> rad.grid.number_of_node_columns
    4
    >>> rad.grid is grid
    True
    >>> np.all(grid.at_cell['radiation__ratio_to_flat_surface'] == 0.)
    True
    >>> np.all(grid.at_node['topographic__elevation'] == 0.)
    True

    >>> grid['node']['topographic__elevation'] = np.array([
    ...       0., 0., 0., 0.,
    ...       1., 1., 1., 1.,
    ...       2., 2., 2., 2.,
    ...       3., 4., 4., 3.,
    ...       4., 4., 4., 4.])
    >>> current_time = 0.5
    >>> rad.update(current_time)
    >>> np.all(grid.at_cell['radiation__ratio_to_flat_surface'] == 0.)
    False
    """

    _name = 'Radiation'

    _input_var_names = (
        'topographic__elevation',
    )

    _output_var_names = (
        'radiation__incoming_shortwave_flux',
        'radiation__ratio_to_flat_surface',
        'radiation__net_shortwave_flux',
    )

    _var_units = {
        'topographic__elevation': 'm',
        'radiation__incoming_shortwave_flux': 'W/m^2',
        'radiation__ratio_to_flat_surface': 'None',
        'radiation__net_shortwave_flux': 'W/m^2',
    }

    _var_mapping = {
        'topographic__elevation': 'node',
        'radiation__incoming_shortwave_flux': 'cell',
        'radiation__ratio_to_flat_surface': 'cell',
        'radiation__net_shortwave_flux': 'cell',
    }

    _var_doc = {
        'topographic__elevation':
            'elevation of the ground surface relative to some datum',
        'radiation__incoming_shortwave_flux':
            'total incident shortwave radiation over the time step',
        'radiation__ratio_to_flat_surface':
            'ratio of total incident shortwave radiation on sloped surface \
             to flat surface',
        'radiation__net_shortwave_flux':
            'net incident shortwave radiation over the time step',
    }

    @use_file_name_or_kwds
    def __init__(self, grid, method='Grid', cloudiness=0.2,
                 latitude=34., albedo=0.2, solar_constant=1366.67,
                 clearsky_turbidity=2., opt_airmass=0., **kwds):
        """
        Parameters
        ----------
        grid: RasterModelGrid
            A grid.
        method: {'Grid'}, optional
            Currently, only default is available.
        cloudiness: float, optional
            Cloudiness.
        latitude: float, optional
            Latitude (radians).
        albedo: float, optional
            Albedo.
        solar_constant: float, optional
            Solar Constant (W/m^2).
        clearsky_turbidity: float, optional
            Clear sky turbidity.
        opt_airmass: float, optional
            Optical air mass.
        """

        self._method = method
        self._N = cloudiness
        self._latitude = latitude
        self._A = albedo
        self._Io = solar_constant
        self._n = clearsky_turbidity
        self._m = opt_airmass

        _assert_method_is_valid(self._method)

        super(Radiation, self).__init__(grid, **kwds)

        for name in self._input_var_names:
            if name not in self.grid.at_node:
                self.grid.add_zeros(name, at='node',
                                    units=self._var_units[name])

        for name in self._output_var_names:
            if name not in self.grid.at_cell:
                self.grid.add_zeros(name, at='cell',
                                    units=self._var_units[name])

        if 'Slope' not in self.grid.at_cell:
            self.grid.add_zeros('Slope', at='cell', units='radians')

        if 'Aspect' not in self.grid.at_cell:
            self.grid.add_zeros('Aspect', at='cell', units='radians')

        self._nodal_values = self.grid['node']
        self._cell_values = self.grid['cell']
        self._slope, self._aspect = \
            grid.calculate_slope_aspect_at_nodes_burrough(
                vals='topographic__elevation')
#        self._slope = grid.calc_slope_of_node( \
#                                elevs = 'topographic__elevation')
#        self._aspect =
        self._cell_values['Slope'] = self._slope
        self._cell_values['Aspect'] = self._aspect

    def update(self, current_time, hour=12., **kwds):
        """Update fields with current loading conditions.

        Parameters
        ----------
        current_time: float
              Current time (years).
        hour: float, optional
              Hour of the day.
        """
        self._t = hour
        self._radf = self._cell_values['radiation__ratio_to_flat_surface']
        self._Rs = self._cell_values['radiation__incoming_shortwave_flux']
        self._Rnet = self._cell_values['radiation__net_shortwave_flux']

        self._julian = np.floor((current_time - np.floor(current_time)) *
                                365.25)    # Julian day

        self._phi = np.radians(self._latitude)    # Latitude in Radians

        self._delta = 23.45 * np.radians(
            np.cos(2*np.pi / 365 * (172 - self._julian)))   # Declination angle

        self._tau = (self._t + 12.0) * np.pi / 12.0     # Hour angle

        self._alpha = np.arcsin(np.sin(self._delta) * np.sin(self._phi) +
                                np.cos(self._delta) * np.cos(self._phi) *
                                np.cos(self._tau))     # Solar Altitude

        if self._alpha <= 0.25 * np.pi / 180.0:    # If altitude is -ve,
            self._alpha = 0.25 * np.pi / 180.0     # sun is beyond the horizon

        # Counting for Albedo, Cloudiness and Atmospheric turbidity
        self._Rgl = (self._Io * np.exp((-1) * self._n * (
            0.128 - 0.054 * np.log10(1. / np.sin(self._alpha)))*(
                1. / np.sin(self._alpha))))

        self._phisun = (np.arctan(- np.sin(self._tau) / (np.tan(self._delta) *
                        np.cos(self._phi) - np.sin(self._phi) *
                        np.cos(self._tau))))      # Sun's Azhimuth
        if (self._phisun >= 0 and - np.sin(self._tau) <= 0):
            self._phisun = self._phisun + np.pi

        elif (self._phisun <= 0 and - np.sin(self._tau) >= 0):
            self._phisun = self._phisun + np.pi

        self._flat = (np.cos(np.arctan(0)) * np.sin(self._alpha) +
                      np.sin(np.arctan(0)) * np.cos(self._alpha) *
                      np.cos(self._phisun - 0))   # flat surface reference

        # flat surface total incoming shortwave radiation
        self._Rsflat = self._Rgl * self._flat

        # flat surface Net incoming shortwave radiation
        self._Rnetflat = ((1 - self._A) * (1 - 0.65 * (self._N ** 2)) *
                          self._Rsflat)

        self._sloped = (np.cos(self._slope) * np.sin(self._alpha) +
                        np.sin(self._slope) * np.cos(self._alpha) *
                        np.cos(self._phisun - self._aspect))
        # Ratio of cosine of solar incidence angle of sloped surface to that
        # of a flat surface
        self._radf = self._sloped / self._flat

        self._radf[self._radf <= 0.] = 0.
        self._radf[self._radf > 6.] = 6.

        # Sloped surface Toatl Incoming Shortwave Radn
        self._Rs = self._Rsflat * self._radf
        self._Rnet = self._Rnetflat * self._radf

        self._cell_values['radiation__ratio_to_flat_surface'] = self._radf
        self._cell_values['radiation__incoming_shortwave_flux'] = self._Rs
        self._cell_values['radiation__net_shortwave_flux'] = self._Rnet
